<?php

/**
 * NETIM Module
 *
 * @package blesta
 * @copyright Copyright (c) 2021, NETIM
 * @link https://www.netim.com NETIM
 */

require_once(__DIR__ . DS . 'lib' . DS . 'NormalizedContact.php');
require_once(__DIR__ . DS . 'lib' . DS . 'Database.php');
require_once(__DIR__ . DS . 'lib' . DS . 'Core.php');

use Netim\NetimAPIException;
use Netim\NormalizedContact;

class Netim extends RegistrarModule
{
	const MODULE_VERSION = '1.2';
    
    /**
     * Initializes the module
     */
	public function __construct()
	{
		# Load language required by this module
		Language::loadLang('netim', null, __DIR__ . DS . 'language' . DS);
        
		# Load config.json
		$this->loadConfig(__DIR__ . DS . 'config.json');
		
		# Load components required by this module
        Loader::loadComponents($this, ['Input', 'Record']);
		
		#Load configuration variables
        Configure::load('netim', dirname(__FILE__) . DS . "config" . DS);
	}
	
	/**
	* Installation module logic goes here.
	* 
	* @see https://docs.blesta.com/display/dev/Modules
	*/
	public function install()
	{
        // Create the netim_async table
        Netim\Database\createAsyncDb($this->Record);
        Netim\Database\createErrorDb($this->Record);

        // Install cron tasks
        Loader::loadModels($this, ['CronTasks']);
        $task = $this->CronTasks->getByKey('netim_async');
 
        if (!$task) {
            $task = [
                'key' => 'netim_async',
                'task_type' => 'module',
                'dir' => 'netim',
                'name' => Language::_( 'Netim.cron_async.name', true),
                'description' => Language::_( 'Netim.cron_async.description', true),
                'type' => 'interval'
            ];
            $task_id = $this->CronTasks->add($task);
        } else
            $task_id = $task->id;

        if ($task_id) {
            $task_run = array(
                'interval' => '10', // Every 5 minutes
                'enabled' => 1, // Enabled by default
            );
            $this->CronTasks->addTaskRun($task_id, $task_run);
        }
    }
	
	/**
	* Upgrade module logic goes here.
	* 
	* @see https://docs.blesta.com/display/dev/Modules
	*/
    public function upgrade($current_version)
	{
        // Ensure new version is greater than installed version
        if (version_compare($this->getVersion(), $current_version) < 0) {
            $this->Input->setErrors(array(
                'version' => array(
                    'invalid' => "Downgrades are not allowed. New version ".$this->getVersion()." vs current version $current_version"
                )
            ));
            return;
        }
    }

	/**
	* Uninstallation module logic goes here.
	* 
	* @see https://docs.blesta.com/display/dev/Modules
	*/
	public function uninstall($module_id, $last_instance)
	{
        // Delete the netim_async table
        Netim\Database\deleteAsyncDb($this->Record);
        Netim\Database\deleteErrorDb($this->Record);

        // Uninstall the cron tasks
        Loader::loadModels($this, ['CronTasks']);
        $task_run = $this->CronTasks->getTaskRunByKey('netim_async','netim',false,'module');
        if($task_run)
            $this->CronTasks->deleteTaskRun($task_run->task_run_id);    
    }
    
    /**
     * Returns the rendered view of the manage module page
     *
     * @param mixed $module A stdClass object representing the module and its accounts
     * @param array $vars An array of post data submitted to or on the manage
     *  module page (used to repopulate fields after an error)
     * @return string HTML content containing information to display when viewing the manager module page
     */
    public function manageModule($module, array &$vars)
    {
        // Load the view into this object, so helpers can be automatically added to the view
        $this->view = new View('manage', 'default');
        $this->view->base_uri = $this->base_uri;
        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'netim' . DS);

        // Load the helpers required for this view
        Loader::loadHelpers($this, array('Form', 'Html', 'Widget'));

        $this->view->set('module', $module);

        return $this->view->fetch();
    }

    /**
     * Returns the rendered view of the add module account page
     *
     * @param array $vars An array of post data submitted to or on the add
     *  module account page (used to repopulate fields after an error)
     * @return string HTML content containing information to display when viewing the add module account page
     */
    public function manageAddRow(array &$vars)
    {
        // Load the view into this object, so helpers can be automatically added to the view
        $this->view = new View('add_row', 'default');
        $this->view->base_uri = $this->base_uri;
        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'netim' . DS);

        // Load the helpers required for this view
        Loader::loadHelpers($this, array('Form', 'Html', 'Widget'));

        // Set unspecified checkboxes
        if (!empty($vars)) {
            $this->setUnspecifiedCheckboxes($vars);
        }
        $this->view->set('vars', (object)$vars);
        return $this->view->fetch();
    }

    /**
     * Returns the rendered view of the edit module account page
     *
     * @param stdClass $module_row The stdClass representation of the existing module account
     * @param array $vars An array of post data submitted to or on the edit
     *  module account page (used to repopulate fields after an error)
     * @return string HTML content containing information to display when viewing the edit module account page
     */
    public function manageEditRow($module_row, array &$vars)
    {
        // Load the view into this object, so helpers can be automatically added to the view
        $this->view = new View('edit_row', 'default');
        $this->view->base_uri = $this->base_uri;
        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'netim' . DS);

        // Load the helpers required for this view
        Loader::loadHelpers($this, array('Form', 'Html', 'Widget'));

        if (empty($vars)) {
            $vars = $module_row->meta;
        } else {
            // Set unspecified checkboxes
            $this->setUnspecifiedCheckboxes($vars);
        }
        $this->view->set('vars', (object)$vars);
        return $this->view->fetch();
    }

    /**
     * Adds the module account. Sets Input errors on failure,
     * preventing the account from being added.
     *
     * @param array $vars An array of module info to add
     * @return array A numerically indexed array of meta fields for the module account containing:
     *  - key The key for this meta field
     *  - value The value for this key
     *  - encrypted Whether or not this field should be encrypted (default 0, not encrypted)
     */
    public function addModuleRow(array &$vars)
    {
        $meta_fields = array('user', 'key', 'sandbox');
        $encrypted_fields = array('key');

        // Set unspecified checkboxes
        $this->setUnspecifiedCheckboxes($vars);

        $this->Input->setRules($this->getRowRules($vars));
  
        // Validate module row
        if ($this->Input->validates($vars)) {
            // Build the meta data for this row
            $meta = array();
            foreach ($vars as $key => $value) {
                if (in_array($key, $meta_fields)) {
                    $meta[] = array(
                        'key' => $key,
                        'value' => $value,
                        'encrypted' => in_array($key, $encrypted_fields) ? 1 : 0
                    );
                }
            }
            return $meta;
        }
    }

    /**
     * Edits the module account. Sets Input errors on failure,
     * preventing the account from being updated.
     *
     * @param stdClass $module_row The stdClass representation of the existing module account
     * @param array $vars An array of module info to update
     * @return array A numerically indexed array of meta fields for the module account containing:
     *  - key The key for this meta field
     *  - value The value for this key
     *  - encrypted Whether or not this field should be encrypted (default 0, not encrypted)
     */
    public function editModuleRow($module_row, array &$vars)
    {
        // Same as adding
        return $this->addModuleRow($vars);
    }

    /**
     * Deletes the module account. Sets Input errors on failure,
     * preventing the account from being deleted.
     *
     * @param stdClass $module_row The stdClass representation of the existing module account
     */
    public function deleteModuleRow($module_row)
    {
    }

    /**
     * Builds and returns the rules required to add/edit a module row
     *
     * @param array $vars An array of key/value data pairs
     * @return array An array of Input rules suitable for Input::setRules()
     */
    private function getRowRules(&$vars)
    {
        return array(
            'user' => array(
                'valid' => array(
                    'rule' => 'isEmpty',
                    'negate' => true,
                    'message' => Language::_('Netim.manage.error.user.valid', true)
                )
            ),
            'key' => array(
                'valid' => array(
                    'last' => true,
                    'rule' => 'isEmpty',
                    'negate' => true,
                    'message' => Language::_('Netim.manage.error.key.valid', true)
                ),
            'valid_connection' => array(
                    'rule' => array(
                        array($this, 'validateConnection'),
                        $vars['user'],
                        $vars['sandbox']
                    ),
                    'message' => Language::_('Netim.manage.error.key.valid_connection', true)
                )
            )
        );
    }

    /**
     * Validates that the given hostname is valid.
     *
     * @param string $host_name The host name to validate
     * @return bool True if the hostname is valid, false otherwise
     */

    public function validateHostName($host_name)
    {
        // Be sure to uncomment the Server use statement at the top of this file if you are going to uncomment this method
        $validator = new Server();
        return $validator->isDomain($host_name) || $validator->isIp($host_name);
    }


    /**
     * Validates that at least 2 name servers are set in the given array of name servers.
     *
     * @param array $name_servers An array of name servers
     * @return bool True if the array count is >= 2, false otherwise
     */
  
    public function validateNameServerCount($name_servers)
    {
        if (is_array($name_servers) && count($name_servers) >= 2) {
            return true;
        }

        return false;
    }

    /**
     * Validates that the nameservers given are formatted correctly.
     *
     * @param array $name_servers An array of name servers
     * @return bool True if every name server is formatted correctly, false otherwise
     */
   
    public function validateNameServers($name_servers)
    {
        // Be sure that you have also uncommented validateHostName() before you uncomment this method
        if (is_array($name_servers)) {
            foreach ($name_servers as $name_server) {
                if (!$this->validateHostName($name_server)) {
                    return false;
                }
            }
        }

        return true;
    }

    /**
     * Validates that the given connection details are correct by attempting to check the availability of a domain
     *
     * @param string $key The API key
     * @param string $user The API user
     * @param string $sandbox "true" if this is a sandbox account, false otherwise
     * @return bool True if the connection details are valid, false otherwise
     */
    public function validateConnection($key, $user, $sandbox)
    {
        try
        {
            $api = $this->getApi(null, $user, $key, $sandbox);
            $api->hello();
            return true;
        }
        catch (Exception $exception)
        {
            return false;
        }
    }

    /**
     * Validates input data when attempting to add a package, returns the meta
     * data to save when adding a package. Performs any action required to add
     * the package on the remote server. Sets Input errors on failure,
     * preventing the package from being added.
     *
     * @param array An array of key/value pairs used to add the package
     * @return array A numerically indexed array of meta fields to be stored for this package containing:
     *
     *  - key The key for this meta field
     *  - value The value for this key
     *  - encrypted Whether or not this field should be encrypted (default 0, not encrypted)
     * @see Module::getModule()
     * @see Module::getModuleRow()
     */
    public function addPackage(array $vars = null)
    {
        // Set rules to validate input data
        $this->Input->setRules($this->getPackageRules($vars));

        // Build meta data to return
        $meta = [];
        if ($this->Input->validates($vars)) {
            if (!isset($vars['meta'] )) {
                return [];
            }

            // Return all package meta fields
            foreach ($vars['meta'] as $key => $value) {
                $meta[] = [
                    'key' => $key,
                    'value' => $value,
                    'encrypted' => 0
                ];
            }
        }

        return $meta;
    }

    /**
     * Validates input data when attempting to edit a package, returns the meta
     * data to save when editing a package. Performs any action required to edit
     * the package on the remote server. Sets Input errors on failure,
     * preventing the package from being edited.
     *
     * @param stdClass $package A stdClass object representing the selected package
     * @param array An array of key/value pairs used to edit the package
     * @return array A numerically indexed array of meta fields to be stored for this package containing:
     *
     *  - key The key for this meta field
     *  - value The value for this key
     *  - encrypted Whether or not this field should be encrypted (default 0, not encrypted)
     * @see Module::getModule()
     * @see Module::getModuleRow()
     */
    public function editPackage($package, array $vars = null)
    {
        // Set rules to validate input data
        $this->Input->setRules($this->getPackageRules($vars));

        // Build meta data to return
        $meta = [];
        if ($this->Input->validates($vars)) {
            if (!isset($vars['meta'] )) {
                return [];
            }

            // Return all package meta fields
            foreach ($vars['meta'] as $key => $value) {
                $meta[] = [
                    'key' => $key,
                    'value' => $value,
                    'encrypted' => 0
                ];
            }
        }

        return $meta;
    }

    /**
     * Builds and returns rules required to be validated when adding/editing a package.
     *
     * @param array $vars An array of key/value data pairs
     * @return array An array of Input rules suitable for Input::setRules()
     */
    private function getPackageRules(array $vars)
    {
        // Validate the package fields
        $rules = [];
        return $rules;
    }

    /**
     * Returns all fields used when adding/editing a package, including any
     * javascript to execute when the page is rendered with these fields.
     *
     * @param $vars stdClass A stdClass object representing a set of post fields
     * @return ModuleFields A ModuleFields object, containg the fields to render
     *  as well as any additional HTML markup to include
     */
    public function getPackageFields($vars = null)
    {              
        Loader::loadHelpers($this, array('Html'));

        $fields = new ModuleFields();

        $types = array(
            'domain' => Language::_('Netim.package_fields.type_domain', true),
        );

        // Set type of package
        $type = $fields->label(Language::_('Netim.package_fields.type', true), 'netim_type');
        $type->attach(
            $fields->fieldSelect(
                'meta[type]',
                $types,
                $this->Html->ifSet($vars->meta['type']),
                array('id' => 'netim_type')
            )
        );
        $fields->setField($type);

        // Set all TLD checkboxes
        $tld_options = $fields->label(Language::_('Netim.package_fields.tlds', true));

        //Getting all tlds that cannot be sold on blesta
        $rejected_tlds = Configure::get('Netim.rejected_tlds');

        // Download TLD data
		$TldData = $this->getNetimRawTldData();
        $tlds = array();
        $tlds_temp = array();

        //Grouping each legacy_prod tlds together
        for($i = 0; $i < count($TldData['extension']); $i++)
        {
            $tld = ((array) $TldData['extension'][$i]);
            if(!isset($tlds_temp[$tld['legacy_prod']]))
                $tlds_temp[$tld['legacy_prod']] = array();
            $arr = $tlds_temp[$tld['legacy_prod']];
            array_push($arr, $tld['@attributes']['name']);
            $tlds_temp[$tld['legacy_prod']] = $arr;
        }

        //Sorting each legacy_prod groups
        foreach($tlds_temp as $legacy => $tldNames)
        {
            sort($tldNames);
            $cpt = 0;
            foreach($tldNames as $tld)
            {
                $tlds[$legacy . $cpt] = ".".$tld;
                $cpt ++;
            }
        }
        
        //Sorting by legacy_prods
        ksort($tlds);
        
        //Turning the temp array containting multiple arrays to a simple array of tlds properly sorted
        foreach ($rejected_tlds as $rejected_tld)
        {
            if(($key = array_search($rejected_tld, $tlds))!== false)
            {
                unset($tlds[$key]);
            }
        }
        foreach ($tlds as $tld) {
            $tld_label = $fields->label($tld, 'tld_' . $tld);
            $tld_options->attach(
                $fields->fieldCheckbox(
                    'meta[tlds][]',
                    $tld,
                    (isset($vars->meta['tlds']) && in_array($tld, $vars->meta['tlds'])),
                    array('id' => 'tld_' . $tld),
                    $tld_label
                    )
                );
            }
            $fields->setField($tld_options);
            
            // Set nameservers
            for ($i=1; $i<=5; $i++) {
                $type = $fields->label(Language::_('Netim.package_fields.ns' . $i, true), 'netim_ns' . $i);
                $type->attach(
                    $fields->fieldText(
                        'meta[ns][]',
                        $this->Html->ifSet($vars->meta['ns'][$i-1]),
                        array('id' => 'netim_ns' . $i)
                        )
                    );
                    $fields->setField($type);
                }
                
                $fields->setHtml("
			<script type=\"text/javascript\">
				$(document).ready(function() {
					toggleTldOptions($('#netim_type').val());

					// Re-fetch module options to pull cPanel packages and ACLs
					$('#netim_type').change(function() {
						toggleTldOptions($(this).val());
					});

					function toggleTldOptions(type) {
						if (type == 'ssl')
							$('.netim_tlds').hide();
						else
							$('.netim_tlds').show();
					}
				});
			</script>
		");

        return $fields;
    }

    /**
     * Adds the service to the remote server. Sets Input errors on failure,
     * preventing the service from being added.
     *
     * @param stdClass $package A stdClass object representing the selected package
     * @param array $vars An array of user supplied info to satisfy the request
     * @param stdClass $parent_package A stdClass object representing the parent
     *  service's selected package (if the current service is an addon service)
     * @param stdClass $parent_service A stdClass object representing the parent
     *  service of the service being added (if the current service is an addon
     *  service and parent service has already been provisioned)
     * @param string $status The status of the service being added. These include:
     *  - active
     *  - canceled
     *  - pending
     *  - suspended
     * @return array A numerically indexed array of meta fields to be stored for this service containing:
     *  - key The key for this meta field
     *  - value The value for this key
     *  - encrypted Whether or not this field should be encrypted (default 0, not encrypted)
     * @see Module::getModule()
     * @see Module::getModuleRow()
     */
    public function addService( $package, array $vars = null, $parent_package = null, $parent_service = null, $status = 'pending')
	{
        Loader::load( __DIR__ . DS . 'api' . DS . 'NormalizedClient.php');
        
        //Nothing to do if the domain is unknown
        if(!isset($vars['domain']))
            return;
        $tld = $this->getTld($vars['domain']);

        $this->_logNetim(__FUNCTION__,[$package,$vars,$parent_package,$parent_service,$status],"input");

        // Manage pending service but already submitted and failed
        // return if addService requested by the cron
        // continue if addService requested by admin with button 'Activate'
        if(Netim\Database\getError($this->Record, 'domainCreate', $vars['domain']))
        {    
            if (isset($vars['save']) && $vars['save']=='Activate')
                Netim\Database\removeError($this->Record, 'domainCreate', $vars['domain']);
            else
            { 
                $this->Input->setErrors(array('api' =>array('errors' => "already in error")));
                return;
            }     
        }

        if(Netim\Database\getError($this->Record, 'domainTransferIn', $vars['domain']))
        {    
            if (isset($vars['save']) && $vars['save']=='Activate')
                Netim\Database\removeError($this->Record, 'domainTransferIn', $vars['domain']);
            else
            { 
                $this->Input->setErrors(array('api' =>array('errors' => "already in error")));
                return;
            }   
        }

        if(Netim\Database\getError($this->Record, 'domainTransferTrade', $vars['domain']))
        {    
            if (isset($vars['save']) && $vars['save']=='Activate')
                Netim\Database\removeError($this->Record, 'domainTransferTrade', $vars['domain']);
            else
            { 
                $this->Input->setErrors(array('api' =>array('errors' => "already in error")));
                return;
            }   
        }

        // Manage Input fields
        $input_fields = [];

        if (isset($vars['transfer']))
        {
            $input_fields = array_merge(
                (array)Configure::get('Netim.transfer_fields'),
                (array)Configure::get('Netim.nameserver_fields'),
                ['Nameservers' => true]
            ); 
        } 
        else 
        {
            $input_fields = array_merge(
                (array)Configure::get('Netim.domain_fields'),
                (array)Configure::get('Netim.nameserver_fields'),
                ['Years' => true, 'Nameservers' => true]
            );

            //Add additional fields
            $additional_fields_settings=Netim\Core\getAdditionalFields($tld);             
            if($additional_fields_settings)
            {
                foreach($additional_fields_settings as $af) 
                    $input_afields[$af->Key] = [];

                $input_fields=array_merge($input_fields,$input_afields);
            } 
        } 

        // Manage pending service
        // return if already submitted at NETIM but request is still pending
        // continue if not submitted
        if(Netim\Database\getAsyncRequest($this->Record,"domainCreate",$vars['domain']) ||
        Netim\Database\getAsyncRequest($this->Record,"domainTransferIn",$vars['domain']) ||
        Netim\Database\getAsyncRequest($this->Record,"domainTransferTrade",$vars['domain'])    
        )
        {
            $this->Input->setErrors(array('api' =>array('errors' => Language::_('Netim.addservice.error.already_pending', true))));
            return;
        }

        // Attempt to validate the input and return on failure
        $this->validateService($package, $vars);
        if ($this->Input->errors())
            return;

        // Process the activation
        if (isset($vars['use_module']) && $vars['use_module'] == 'true') 
        {
            if ($package->meta->type == 'domain')
            {
               	// Init basic information
				if (!isset($this->Clients)) {
					Loader::loadModels($this, array('Clients'));
				}
				if (!isset($this->Contacts)) {
					Loader::loadModels($this, array('Contacts'));
				}
				
				$client = $this->Clients->get($vars['client_id']);
				$numbers = $this->Contacts->getNumbers($client->contact_id, 'phone');             
                
                // Check Trustee and local contact service existance for that tld
                $trustee = (isset($vars["configoptions"]["Trustee service"]) && $vars["configoptions"]["Trustee service"] == "TRUSTEE") || 
                    in_array($tld,Configure::get('Netim.forcedtrustee'));
                $localContactService = (isset($vars["configoptions"]["Local contact service"]) && $vars["configoptions"]["Local contact service"] == "LOC") ||
                    in_array($tld,Configure::get('Netim.forcedlocalcontact'));

                try
                {
                    //get API
                    $api = $this->getApi($this->getModuleRow($package->module_row));

                    if($trustee || $localContactService)
                        $tldInfo = $api->domainTldInfo($tld);
                }
                catch (NetimAPIException $exception)
                {
                    $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                }
                finally
                {
                    $this->_logAPIRequest(__FUNCTION__,$api);
                }

                if($trustee && $tldInfo->HasTrusteeService == "0")
                {
                    $this->Input->setErrors(array('api' =>array('errors' => Language::_('Netim.addservice.error.trustee', true))));
                    return;
                } 

                if($localContactService && $tldInfo->HasLocalContactService == "0")
                {
                    $this->Input->setErrors(array('api' =>array('errors' => Language::_('Netim.addservice.error.localcontact', true))));
                    return;
                } 

                // Normalize contact information to be used with API
                $contact = new NormalizedContact(
                        $client->first_name, $client->last_name, $client->company, $client->address1, 
                        $client->address2, $client->zip, $client->state, $client->country, $client->city, ((isset($numbers[0])) ? $numbers[0]->number : null),
                        $client->email, 'EN', 1);
                
                // Manage additional domain fields
                $this->parseAdditionalfields($vars,$tld,$contact);

                // Define or create ID Contact to be used
                try
                {
                    // Trustee service or client
                    if($trustee)
                    {
                        $idOwner="";
                        $trusteeArray = Configure::get('Netim.trusteeIDs');
                        foreach($trusteeArray as $key=>$value)
                        {
                            if(in_array($tld,$value))
                                 $idOwner=$key;
                        }

                        if(empty($idOwner))
                        {
                            $error=Language::_('Netim.!error.no_trustee_id', true);
                            $this->Input->setErrors(array('api' =>array('errors' => $error)));
                            return;
                        }
                    }
                    else
                        $idOwner = $api->contactCreateObj($contact);
                }
                catch(NetimAPIException $exception)
                {
                    $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage()))); 
                    return;
                }
                finally
                {
                    $this->_logAPIRequest(__FUNCTION__,$api);
                }

                try{    
                    // Local contact service or client
                    $contact->setIsOwner(false);
                    $idContact = $api->contactCreateObj($contact);

                    if($localContactService)
                    {
                        $idAdmin="";
                        $localContactArray = Configure::get('Netim.localcontactIDs');
                        foreach ($localContactArray as $key => $value)
                        {
                            if(in_array($tld,$value))
                                $idAdmin=$key;
                        }

                        if(empty($idAdmin))
                        {
                            $error=Language::_('Netim.!error.no_local_contact_id', true);
                            $this->Input->setErrors(array('api' =>array('errors' => $error)));
                            return;
                        }
                    }
                    else
                        $idAdmin = $idContact;  
                }
                catch (NetimAPIException $exception)
                {
                    $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                    return;
                }
                finally
                {
                    $this->_logAPIRequest(__FUNCTION__,$api);
                }
				
                // ******************
                // domain transfer in
                // ******************    
                if (isset($vars['transfer'])) 
                {
                    try
                    {
                        // if Trustee, should we request a domainTransferTrade
                        if($trustee)
                        {
                            $opeInfo = $api->QueryOpeList($tld);
                            if($opeInfo->domainTransferTrade==1)
                                $request_type="domainTransferTrade";
                            else
                                $request_type="domainTransferIn"; 
                        }
                        else
                            $request_type="domainTransferIn"; 

                        if(!isset($vars['transfer_key']) || $vars['transfer_key']==false)
                            $vars['transfer_key']="";

                        // Request a domain transfer with change of domain holder to Trustee  
                        if($request_type=="domainTransferTrade")    
                            $result = $api->domainTransferTrade($vars['domain'], $vars['transfer_key'], $idOwner, $idAdmin, $idContact, $idContact, $vars["ns1"], $vars["ns2"], $vars["ns3"], $vars["ns4"], $vars["ns5"]);
                        else // Request a regular domain transfer  
                            $result = $api->domainTransferIn($vars['domain'], $vars['transfer_key'], $idOwner, $idAdmin, $idContact, $idContact, $vars["ns1"], $vars["ns2"], $vars["ns3"], $vars["ns4"], $vars["ns5"]);
                        $this->_logAPIRequest(__FUNCTION__,$api);
                    }
                    catch (NetimAPIException $exception)
                    {
                        Netim\Database\addError($this->Record, $exception->getMessage(), $request_type, $vars['domain']);
                        $this->_logAPIRequest(__FUNCTION__,$api);
                        $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                        return;
                    }

                    if ($result->STATUS == "Done") {
                        //Nothing to do here as this case never happens
                    }

                    if ($result->STATUS == "Pending")
                    {
                        // Record the domain as asynchronous pending request
                        Netim\Database\addAsyncRequest($this->Record, $result->ID_OPE,  $result->TYPE, $vars['domain']);

                        $message=str_replace("[ID_OPE]",$result->ID_OPE,Language::_('Netim.addservice.error.transfer_pending', true));
                        $this->Input->setErrors(array('api' =>array('errors' => $message)));
                        return;
                    }
        
                    if ($result->STATUS == "Failed")
                    {
                        // Record the domain as failed
                        Netim\Database\addError($this->Record, $result->MESSAGE, $request_type, $result->MESSAGE);

                        $this->Input->setErrors(array('api' =>array('errors' => $result->MESSAGE)));
                        return;
                    }
                } 
                else 
                {
                    // *******************
                    // domain registration
                    // *******************

                    // Get Duration
                    foreach($package->pricing as $term)
                    {
                        if($term->id == $vars["pricing_id"])
                            $duration = $term->term;
                    }

                    // Request domain registration
                    try
                    {
                        $result = $api->domainCreate($vars['domain'], $idOwner, $idAdmin, $idContact, $idContact, $vars["ns1"], $vars["ns2"], $vars["ns3"], $vars["ns4"], $vars["ns5"], $duration);
                    }
                    catch (NetimAPIException $exception)
                    {
                        Netim\Database\addError($this->Record, $exception->getMessage(), 'domainCreate', $vars['domain']);
                        $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                        return;
                    }
                    finally
                    {
                        $this->_logAPIRequest(__FUNCTION__,$api);                    
                    }
                    
                    if ($result->STATUS == "Done") {
                        //Nothing to do here as this case never happens
                    }

                    if ($result->STATUS == "Pending")
                    {
                        // Record the domain as asynchronous pending request
                        Netim\Database\addAsyncRequest($this->Record, $result->ID_OPE, "domainCreate", $vars['domain']);

                        $message=str_replace("[ID_OPE]",$result->ID_OPE,Language::_('Netim.addservice.error.registration_pending', true));
                        $this->Input->setErrors(array('api' =>array('errors' => $message)));
                        return;
                    }
                    
                    if ($result->STATUS == "Failed")
                    {
                        // Record the domain as failed
                        Netim\Database\addError($this->Record, $result->MESSAGE, 'domainCreate', $vars['domain']);
                        $this->Input->setErrors(array('api' =>array('errors' => $result->MESSAGE)));
                        return;
                    }                
                }
            }
        }

        //Default return
        $meta = array();
        $fields = array_intersect_key($vars, $input_fields);
        foreach ($fields as $key => $value) {
            $meta[] = array(
                'key' => $key,
                'value' => $value,
                'encrypted' => 0
            );
        }

        $this->_logNetim(__FUNCTION__,$meta,"output");

        return $meta;
    }

    function parseAdditionalfields($vars,$tld,&$contact)
	{
        //Parse all $vars into additional fields
        $contact->setCompanyNumber(isset($vars["companyNumber"]) ? $vars["companyNumber"] : "");
        $contact->setTmName(isset($vars["tmName"]) ? $vars["tmName"] : "");
        $contact->setTmNumber(isset($vars["tmNumber"]) ? $vars["tmNumber"] : "");
        $contact->setTmType(isset($vars["tmType"]) ? $vars["tmType"] : "");
        $contact->setTmDate(isset($vars["tmDate"]) ? $vars["tmDate"] : "");
        $contact->setVatNumber(isset($vars["vatNumber"]) ? $vars["vatNumber"] : "");
        $contact->setBirthDate(isset($vars["birthDate"]) ? $vars["birthDate"] : "");
        $contact->setBirthZipCode(isset($vars["birthZipCode"]) ? $vars["birthZipCode"] : "");
        $contact->setBirthCity(isset($vars["birthCity"]) ? $vars["birthCity"] : "");
        $contact->setBirthCountry(isset($vars["birthCountry"]) ? $vars["birthCountry"] : "");
        $contact->setIdNumber(isset($vars["idNumber"]) ? $vars["idNumber"] : "");

		//If the companyName exists, we replace the bodyName
        if(isset($vars["companyName"]) && !empty($vars["companyName"]))
            $contact->setBodyName($vars["companyName"]);

		//If the firstName exists, we replace the firstName
		if(isset($vars["firstName"]) && !empty($vars["firstName"]))
            $contact->setFirstName($vars["firstName"]);

		//If the lastName exists, we replace the bodyName
		if(isset($vars["lastName"]) && !empty($vars["lastName"]))
			$contact->setLastName($vars["lastName"]);
		
        //Special case for .ru
        if(isset($vars["RU_TIN"]) || isset($vars["RU_KPP"]))
        {
            $contact->setCompanyNumber($vars["RU_TIN"]."#".$vars["RU_KPP"]);
            unset($vars["RU_TIN"]);
            unset($vars["RU_KPP"]);
        }     

		//Additional fields
        $additional=array();
        $additional_fields_settings=Netim\Core\getAdditionalFields($tld);
        
        if($additional_fields_settings)
        {
            foreach($additional_fields_settings as $field) 
            {
                if(isset($vars[$field->Key]) && preg_match("#_#",$field->Key))
                {
                    if(strpos($field->Key, '|') !== false)    
                    {
                        $k=strstr($field->Key, '|', true);
                        $additional[$k][$vars['domain']]=$vars[$field->Key];
                    }
                    else
                        $additional[$field->Key]=$vars[$field->Key];    
                }   
            }     
        }

        $contact->setAdditional($additional);
	}

    /**
     * Edits the service on the remote server. Sets Input errors on failure,
     * preventing the service from being edited.
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $vars An array of user supplied info to satisfy the request
     * @param stdClass $parent_package A stdClass object representing the parent
     *  service's selected package (if the current service is an addon service)
     * @param stdClass $parent_service A stdClass object representing the parent
     *  service of the service being edited (if the current service is an addon service)
     * @return array A numerically indexed array of meta fields to be stored for this service containing:
     *  - key The key for this meta field
     *  - value The value for this key
     *  - encrypted Whether or not this field should be encrypted (default 0, not encrypted)
     * @see Module::getModule()
     * @see Module::getModuleRow()
     */
    public function editService($package, $service, array $vars = array(), $parent_package = null, $parent_service = null)
    {
        // All is handled by admin/client tabs instead
        return null; 
    }

    /**
     * Suspends the service on the remote server. Sets Input errors on failure,
     * preventing the service from being suspended.
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param stdClass $parent_package A stdClass object representing the parent
     *  service's selected package (if the current service is an addon service)
     * @param stdClass $parent_service A stdClass object representing the parent
     *  service of the service being suspended (if the current service is an addon service)
     * @return mixed null to maintain the existing meta fields or a numerically
     *  indexed array of meta fields to be stored for this service containing:
     *  - key The key for this meta field
     *  - value The value for this key
     *  - encrypted Whether or not this field should be encrypted (default 0, not encrypted)
     * @see Module::getModule()
     * @see Module::getModuleRow()
     */
    public function suspendService($package, $service, $parent_package = null, $parent_service = null)
    {
        // Not available from reseller level
        // Reseller should update the domain to its trash nameservers or remove all nameservers
        $this->Input->setErrors(array('api' =>array('errors' => Language::_('Netim.error.not_implemented', true))));
        return;
    }

    /**
     * Unsuspends the service on the remote server. Sets Input errors on failure,
     * preventing the service from being unsuspended.
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param stdClass $parent_package A stdClass object representing the parent
     *  service's selected package (if the current service is an addon service)
     * @param stdClass $parent_service A stdClass object representing the parent
     *  service of the service being unsuspended (if the current service is an addon service)
     * @return mixed null to maintain the existing meta fields or a numerically
     *  indexed array of meta fields to be stored for this service containing:
     *  - key The key for this meta field
     *  - value The value for this key
     *  - encrypted Whether or not this field should be encrypted (default 0, not encrypted)
     * @see Module::getModule()
     * @see Module::getModuleRow()
     */
    public function unsuspendService($package, $service, $parent_package = null, $parent_service = null)
    {
        // Not available from reseller level
        // Reseller should update the domain to initial nameservers
        $this->Input->setErrors(array('api' =>array('errors' => Language::_('Netim.error.not_implemented', true))));
        return;
    }

     /**
     * Cancels the service on the remote server. Sets Input errors on failure,
     * preventing the service from being canceled.
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param stdClass $parent_package A stdClass object representing the parent
     *  service's selected package (if the current service is an addon service)
     * @param stdClass $parent_service A stdClass object representing the parent
     *  service of the service being canceled (if the current service is an addon service)
     * @return mixed null to maintain the existing meta fields or a numerically
     *  indexed array of meta fields to be stored for this service containing:
     *  - key The key for this meta field
     *  - value The value for this key
     *  - encrypted Whether or not this field should be encrypted (default 0, not encrypted)
     * @see Module::getModule()
     * @see Module::getModuleRow()
     */
    public function cancelService($package, $service, $parent_package = null, $parent_service = null)
    {
        $this->_logNetim(__FUNCTION__,[$package,$service,$parent_package,$parent_service],"input");

        Loader::loadModels($this, array("Services"));
        $domain = $this->getServiceDomain($service);
        $tld = $this->getTld($domain);

        // Download TLD data
		$TldData = $this->getNetimRawTldData();
        if(isset($TldData["error"]))
        {
            $this->Input->setErrors(array('api' =>array('errors' => $TldData["error"])));
            return;
        }

        for($i=0;$i<count($TldData['extension']);$i++)
        {
            if(preg_match('#.'.$tld.'#',$TldData['extension'][$i]['@attributes']['name']))
            {
                $id = $i;
                break;
            }
        }

        if($TldData['extension'][$i]['HasImmediateDelete'])
        {
            try
            {
                //get API
                $api = $this->getApi($this->getModuleRow($package->module_row));

                //delete Domain
                $response = $api->domainDelete($domain, 'NOW');
                if ($response->STATUS == 'Done')
                {
                    if (!$this->Services->edit($service->id, ['status' => "canceled"],true,true));
                        $this->Input->setErrors(array('api' =>array('errors' => "Domain has been deleted at the Registrar but the cancelation in Blesta failed")));
                }
                if ($response->STATUS == 'Pending')
                {
                    Netim\Database\addAsyncRequest($this->Record, $response->ID_OPE, 'domainDelete', $domain);
                    $this->Input->setErrors(array('api' =>array('errors' => "Operation is pending with id " . $response->ID_OPE)));
                }
                if ($response->STATUS == 'Failed')
                {
                    $this->Input->setErrors(array('api' =>array('errors' => "Operation failed : " . $response->MESSAGE)));
                }
            }
            catch (NetimAPIException $exception)
            {
                $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
            }
            finally
            {
                $this->_logAPIRequest(__FUNCTION__,$api);
            }
        }
        else
        {
            try
            {
                //get API
                $api = $this->getApi($this->getModuleRow($package->module_row));

                $result  = $api->domainSetAutoRenew($this->getServiceDomain($service), "0");
                if($TldData['extension'][$i]['MinimumDNS'] == 0)
                    $response = $api->domainChangeDNS($domain);
                else
                    $response = $api->domainChangeDNS($domain,"ns1-trash.netim.net","ns2-trash.netim.net");
                $this->Services->edit($service->id, ['status' => "canceled"],true,true);
            }
            catch (NetimAPIException $exception)
            {
                $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                return;
            }
            finally
            {
                $this->_logAPIRequest(__FUNCTION__,$api);
            }
        }

        $this->_logNetim(__FUNCTION__,null,"output");

        return null;
    }

    /**
     * Allows the module to perform an action when the service is ready to renew.
     * Sets Input errors on failure, preventing the service from renewing.
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param stdClass $parent_package A stdClass object representing the parent
     *  service's selected package (if the current service is an addon service)
     * @param stdClass $parent_service A stdClass object representing the parent
     *  service of the service being renewed (if the current service is an addon service)
     * @return mixed null to maintain the existing meta fields or a numerically
     *  indexed array of meta fields to be stored for this service containing:
     *  - key The key for this meta field
     *  - value The value for this key
     *  - encrypted Whether or not this field should be encrypted (default 0, not encrypted)
     * @see Module::getModule()
     * @see Module::getModuleRow()
     */
    public function renewService($package, $service, $parent_package = null, $parent_service = null)
    {
        $this->_logNetim(__FUNCTION__,[$package,$service,$parent_package,$parent_service],"input");

        Loader::loadModels($this, array("Services"));

        try {
            //get API
            $api = $this->getApi($this->getModuleRow($package->module_row));
            
            $result  = $api->domainRenew($this->getServiceDomain($service), $service->package_pricing->term);
        }
        catch(NetimAPIException $fault) {
        }
        finally {
            $this->_logAPIRequest(__FUNCTION__,$api);
        }

        try
        {
            //Get and Set the expiration date after renewal attempt
            //If result is done, the domain gets the real expiration date after renew
            //If result is pending, the domain gets back the current expiration date until the cron updates it
            //If result is failed, the domain gets back the real expiration date.
            $domainInfo = $api->domainInfo($this->getServiceDomain($service));
            $this->Services->edit($service->id, ['date_renews' => $domainInfo->dateExpiration],true,false); 
        }
        catch(NetimAPIException $fault) {
        }
        finally {
            $this->_logAPIRequest(__FUNCTION__,$api);
        }
                    
        if($result && $result->STATUS=="Pending")
        {
            // Record the domain as asynchronous pending request
            Netim\Database\addAsyncRequest($this->Record, $result->ID_OPE, 'domainRenew', $this->getServiceDomain($service));
        }

        $this->_logNetim(__FUNCTION__,null,"output");

        return null;
    }

    /**
     * Attempts to validate service info. This is the top-level error checking method. Sets Input errors on failure.
     *
     * @param stdClass $package A stdClass object representing the selected package
     * @param array $vars An array of user supplied info to satisfy the request
     * @return bool True if the service validates, false otherwise. Sets Input errors when false.
     */
    public function validateService($package, array $vars = null)
    {  
        $this->_logNetim(__FUNCTION__,[$package,$vars],"input");

        $user_lang = Configure::get('Blesta.language');
        $rules = [];

        if ($package->meta->type == 'domain') 
        {
            // Validate the domain name
            $r = [
                'rule' => ['isEmpty'],
                'negate' => true,
                'message' => Language::_('Netim.domain.error', true),
                'post_format' => 'trim'
            ];

            $rule['domain']['empty']=$r;
            $rules = array_merge($rules, $rule);
        
           if (isset($vars['transfer']) && isset($vars['transfer_key']) && empty($vars['transfer_key']))
            {
                // Validate the transfer Key in case of domain transfer
                $r = [
                    'rule' => ['isEmpty'],
                    'negate' => true,
                    'message' => Language::_('Netim.transfer.error.transfer_key', true),
                    'post_format' => 'trim'
                ];

                $rule['transfer_key']['empty']=$r;
                $rules = array_merge($rules, $rule);
            }
            else
            {
                // If trustee service exist and is not selected, we apply the validation rules. Otherwise, nothing to check
                $validation = (isset($vars["configoptions"]["Trustee service"]) && $vars["configoptions"]["Trustee service"] == "TRUSTEE") ||
                (isset($vars["configoptions"]["Local contact service"]) && $vars["configoptions"]["Local contact service"] == "LOC");
                if(!$validation)
                {
                    // Validate additional domain fields
                    $tld = $this->getTld($vars['domain']);
                    $additional_fields_settings=Netim\Core\getAdditionalFields($tld);
                    
                    if($additional_fields_settings)
                    {
                        //Build the rules
                        $display_rules=false;
                        foreach($additional_fields_settings as $af) 
                        {
                            $rule[$af->Key] = [];

                            //Get the translated name
                            $label=Netim\Core\translate($af->Name,$user_lang);
    
                            //Required rule
                            $required = (isset($af->Options->Required) && $af->Options->Required);
                            if($af->Key=="rules") // Display rules is considered as a checkbox and is mandatory
                            {
                                $display_rules=true;
                                $required=true;
                            }     

                            if($af->Key=="notices") // Display notices is considered as a checkbox and is mandatory
                                 $required=true;

                            if($af->Key=="agreement" && $display_rules) //As display rules is considered as an agreement, checkbox agreement is not taken into account
                                continue;

                            foreach($additional_fields_settings as $af2) 
                            {
                                $required = $required || (isset($af2->Options->Requires) && in_array($af->Id,$af2->Options->Requires) && !empty($vars[$af2->Key]));

                                if(isset($af2->Items))
                                    foreach($af2->Items as $item)
                                        $required = $required || (isset($item->Requires) && in_array($af->Id,$item->Requires) && !empty($vars[$af2->Key]) && $vars[$af2->Key]==$item->Value);    
                            }

                            if($required)
                            {
                                if($af->Type=="display" || $af->Type=="checkbox")
                                    $message=Language::_('Netim.validate.notchecked', true, $label);
                                else
                                    $message=Language::_('Netim.validate.empty', true, $label);
                                
                                $r = [
                                    'rule' => ['isEmpty'],
                                    'negate' => true,
                                    'message' => $message,
                                    'post_format' => 'trim'
                                ];

                                $rule[$af->Key]['empty']=$r;
                            }

                            //Syntax rule
                            if(isset($af->Options->Syntax) && $af->Options->Syntax && !empty($vars[$af->Key]))
                            {
                                $r = [
                                    'rule' => ['matches', '/^'.$af->Options->Syntax.'$/'],
                                    'if_set' => true,
                                    'message' => Language::_('Netim.validate.wrong_syntax', true, $label),
                                    'post_format' => 'trim'
                                ];

                                $rule[$af->Key]['format']=$r;
                            }

                            if(count($rule[$af->Key])>0)
                                $rules = array_merge($rules, $rule);
                        }
                    }
                }
            }
        }
  
        if (isset($rules) && count($rules) > 0) {
            $this->Input->setRules($rules);
            $this->_logNetim(__FUNCTION__,$this->Input->validates($vars,"output"),"");
            return $this->Input->validates($vars);
        }

        $this->_logNetim(__FUNCTION__,"true","output");

        return true;      
    }

    /**
     * Fetches the HTML content to display when viewing the service info in the
     * admin interface.
     *
     * @param stdClass $service A stdClass object representing the service
     * @param stdClass $package A stdClass object representing the service's package
     * @return string HTML content containing information to display when viewing the service info
     */
    public function getAdminServiceInfo($service, $package)
    {
        // --> Not used anymore in domain manager plugin in admin area <--
        return $this->manageServiceInfo('service_info_row', $service, $package);
    }

    /**
     * Fetches the HTML content to display when viewing the service info in the
     * client interface.
     *
     * @param stdClass $service A stdClass object representing the service
     * @param stdClass $package A stdClass object representing the service's package
     * @return string HTML content containing information to display when viewing the service info
     */
    public function getClientServiceInfo($service, $package)
    {
         return $this->manageServiceInfo('client_service_info_row', $service, $package);
    }

    /**
     * Fetches the HTML content to display when viewing the service info in the
     * client interface.
     *
     * @param string $view The view to use
     * @param stdClass $service A stdClass object representing the service
     * @param stdClass $package A stdClass object representing the service's package
     * @return string HTML content containing information to display when viewing the service info
     */
    private function manageServiceInfo($view, $service, $package)
    {
        $this->_logNetim(__FUNCTION__,[$view,$service,$package],"input");

        $this->view = new View($view, 'default');

        //Get domain info
        if($service->status!="pending")
        {
            try 
            {
                //get API
                $api = $this->getApi($this->getModuleRow($package->module_row));
                
                $info = $api->domainInfo($this->getServiceDomain($service));
                $this->view->set("info", $info);  
            }
            catch (NetimAPIException $exception)
            {
                $this->view->set("info", false);
            }
            finally
            {
                $this->_logAPIRequest(__FUNCTION__,$api);
            }
        }

        $error = null;
        if($errorCreate=(Netim\Database\getError($this->Record, 'domainCreate', $this->getServiceDomain($service))))
            $error = $errorCreate;
        if($errorTransferIn=(Netim\Database\getError($this->Record, 'domainTransferIn', $this->getServiceDomain($service))))
            $error = $errorTransferIn;
        if($errorTransferTrade=(Netim\Database\getError($this->Record, 'domainTransferTrade', $this->getServiceDomain($service))))
            $error = $errorTransferTrade;

        $this->view->set("status", $service->status);
        $this->view->set("domain", $this->getServiceDomain($service));
        $this->view->set("tld", $this->getTld($this->getServiceDomain($service)));
        $this->view->set("idn", \idn_to_ascii($this->getServiceDomain($service), IDNA_NONTRANSITIONAL_TO_ASCII));
        if($error)
            $this->view->set("error", $error);
        $asyncRequests = Netim\Database\getAllAsyncRequests($this->Record);
        $requests= [];
        foreach ($asyncRequests as $request)
        {
            if($request->domaine == $this->getServiceDomain($service))
                $requests[]=$request;
        }
        $this->view->set("requests", $requests);
        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'netim' . DS);
        $return = $this->view->fetch();

        $this->_logNetim(__FUNCTION__,$return,"output");

        return $return;
    }

    /**
     * Returns all tabs to display to a client when managing a service.
     *
     * @param stdClass $service A stdClass object representing the service
     * @return array An array of tabs in the format of method => title, or method => array where array contains:
     *
     *  - name (required) The name of the link
     *  - icon (optional) use to display a custom icon
     *  - href (optional) use to link to a different URL
     *      Example:
     *      ['methodName' => "Title", 'methodName2' => "Title2"]
     *      ['methodName' => ['name' => "Title", 'icon' => "icon"]]
     */
    public function getClientServiceTabs($service)
    {
        //$this->_logNetim(__FUNCTION__,$service,"input");

        $array = array();
        $array['tabClientWhois'] = Language::_('Netim.tab_whois.title', true);
        $array['tabClientNameservers'] = Language::_('Netim.tab_nameservers.title', true);
        $array['tabClientSettings'] = Language::_('Netim.tab_settings.title', true);
        $array['tabClientDnsrecord'] = Language::_('Netim.tab_dnsrecord.title', true);
        $array['tabClientHosts'] = Language::_('Netim.tab_hosts.title', true);
        $array['tabClientMails'] = Language::_('Netim.tab_mails.title',true);
        
        // Check if DNS Management is enabled
        if (!$this->featureServiceEnabled('dns_management', $service)) {
            unset($array['tabClientDnsrecord']);
        }

        // Check if Email Forwarding is enabled
        if (!$this->featureServiceEnabled('email_forwarding', $service)) {
            unset($array['tabClientMails']);
        }

        //$this->_logNetim(__FUNCTION__,$array,"output");
        
        return $array;
    }

    /**
     * Returns all tabs to display to an admin when managing a service
     *
     * @param stdClass $service A stdClass object representing the service
     * @return array An array of tabs in the format of method => title.
     *  Example: ['methodName' => "Title", 'methodName2' => "Title2"]
     */
    public function getAdminServiceTabs($service)
    {        
        //$this->_logNetim(__FUNCTION__,$service,"input");

        $array = array();
        $array['tabWhois'] = Language::_('Netim.tab_whois.title', true);
        $array['tabNameservers'] = Language::_('Netim.tab_nameservers.title', true);
        $array['tabSettings'] = Language::_('Netim.tab_settings.title', true);
        $array['tabDnsrecord'] = Language::_('Netim.tab_dnsrecord.title', true);
        $array['tabDnsSec'] = Language::_('Netim.tab_dnssec.title', true);
        $array['tabHosts'] = Language::_('Netim.tab_hosts.title', true);
        $array['tabMails'] = Language::_('Netim.tab_mails.title',true);

        
        // Check if DNS Management is enabled
        if (!$this->featureServiceEnabled('dns_management', $service)) {
            unset($array['tabDnsrecord']);
        }

        // Check if Email Forwarding is enabled
        if (!$this->featureServiceEnabled('email_forwarding', $service)) {
            unset($array['tabMails']);
        }
        
        $array['tabCost'] = Language::_('Netim.tab_cost.title', true);

        //$this->_logNetim(__FUNCTION__,$array,"output");

        return $array;
        
    }

    /**
     * Checks if a feature is enabled for a given service
     *
     * @param string $feature The name of the feature to check if it's enabled (e.g. id_protection)
     * @param stdClass $service An object representing the service
     * @return bool True if the feature is enabled, false otherwise
     */
    private function featureServiceEnabled($feature, $service)
    {
        // Get service option groups
        foreach ($service->options as $option) {
            if ($option->option_name == $feature) {
                return true;
            }
        }

        return false;
    }

    /**
     * Admin Whois tab
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $get Any GET parameters
     * @param array $post Any POST parameters
     * @param array $files Any FILES parameters
     * @return string The string representing the contents of this tab
     */
    public function tabWhois($package, $service, array $get = null, array $post = null, array $files = null)
    {
         return $this->manageWhois('tab_whois', $package, $service, $get, $post, $files);
    }

    /**
     * Client Whois tab
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $get Any GET parameters
     * @param array $post Any POST parameters
     * @param array $files Any FILES parameters
     * @return string The string representing the contents of this tab
     */
    public function tabClientWhois($package, $service, array $get = null, array $post = null, array $files = null)
    {
        return $this->manageWhois('tab_client_whois', $package, $service, $get, $post, $files);
    }

    /**
     * Admin Nameservers tab
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $get Any GET parameters
     * @param array $post Any POST parameters
     * @param array $files Any FILES parameters
     * @return string The string representing the contents of this tab
     */
    public function tabNameservers($package, $service, array $get = null, array $post = null, array $files = null)
    {
        return $this->manageNameservers('tab_nameservers', $package, $service, $get, $post, $files);
    }

    /**
     * Admin Nameservers tab
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $get Any GET parameters
     * @param array $post Any POST parameters
     * @param array $files Any FILES parameters
     * @return string The string representing the contents of this tab
     */
    public function tabClientNameservers($package, $service, array $get = null, array $post = null, array $files = null)
    {
        return $this->manageNameservers('tab_client_nameservers', $package, $service, $get, $post, $files);
    }

    /**
     * Admin Cost pricing tab
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $get Any GET parameters
     * @param array $post Any POST parameters
     * @param array $files Any FILES parameters
     * @return string The string representing the contents of this tab
     */
    public function tabCost($package, $service, array $get = null, array $post = null, array $files = null)
    {
        $this->_logNetim(__FUNCTION__,[$package,$service,$get,$post,$files],"input");

        $this->view = new View('tab_cost', 'default');
        $row = $this->getModuleRow($package->module_row);
        $fields = $this->serviceFieldsToObject($service->fields);
        
        try
        {
            //get API
            $api = $this->getApi($this->getModuleRow($package->module_row));

            $cost = $api->queryDomainPrice($fields->domain);
            $list = $api->queryOpeList($this->getTld($domain));
            $tldInfo = $api->domainTldInfo($this->getTld($domain));

            $this->view->set('cost', $cost);
            $this->view->set('list', $list);
            $this->view->set('tldInfo', $tldInfo);
        }
        catch (NetimAPIException $exception)
        {
            $this->_logAPIRequest(__FUNCTION__,$api);
        }

        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'netim' . DS);
        $return = $this->view->fetch();
        $this->_logNetim(__FUNCTION__,$return,"output");
        return $return;
    }

    /**
     * Admin Settings tab
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $get Any GET parameters
     * @param array $post Any POST parameters
     * @param array $files Any FILES parameters
     * @return string The string representing the contents of this tab
     */
    public function tabSettings($package, $service, array $get = null, array $post = null, array $files = null)
    {
        return $this->manageSettings('tab_settings', $package, $service, $get, $post, $files);
    }

    /**
     * Client Settings tab
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $get Any GET parameters
     * @param array $post Any POST parameters
     * @param array $files Any FILES parameters
     * @return string The string representing the contents of this tab
     */
    public function tabClientSettings($package, $service, array $get = null, array $post = null, array $files = null)
    {
        return $this->manageSettings('tab_client_settings', $package, $service, $get, $post, $files);
    }

    /**
     * Admin DNS record tab
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $get Any GET parameters
     * @param array $post Any POST parameters
     * @param array $files Any FILES parameters
     * @return string The string representing the contents of this tab
     */
    public function tabDnsrecord($package, $service, array $get = null, array $post = null, array $files = null)
    {
        return $this->manageDnsRecords('tab_dnsrecords', $package, $service, $get, $post, $files);
    }

    /**
     * Client DNS record tab
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $get Any GET parameters
     * @param array $post Any POST parameters
     * @param array $files Any FILES parameters
     * @return string The string representing the contents of this tab
     */
    public function tabClientDnsRecord($package, $service, array $get = null, array $post = null, array $files = null)
    {
        return $this->manageDnsRecords('tab_client_dnsrecords', $package, $service, $get, $post, $files);
    }

    public function tabDnssec($package, $service, array $get = null, array $post = null, array $files = null)
    {
        return $this->manageDnssec('tab_dnssec', $package, $service, $get, $post, $files);
    }

    /**
     * Admin Host tab
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $get Any GET parameters
     * @param array $post Any POST parameters
     * @param array $files Any FILES parameters
     * @return string The string representing the contents of this tab
     */
    public function tabHosts($package, $service, array $get = null, array $post = null, array $files = null)
    {
        return $this->manageHosts('tab_hosts', $package, $service, $get, $post, $files);
    }

    /**
     * Client Host tab
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $get Any GET parameters
     * @param array $post Any POST parameters
     * @param array $files Any FILES parameters
     * @return string The string representing the contents of this tab
     */
    public function tabClientHosts($package, $service, array $get = null, array $post = null, array $files = null)
    {
        return $this->manageHosts('tab_client_hosts', $package, $service, $get, $post, $files);
    }

    /**
     * Admin Mail redirection tab
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $get Any GET parameters
     * @param array $post Any POST parameters
     * @param array $files Any FILES parameters
     * @return string The string representing the contents of this tab
     */
    public function tabMails($package, $service, array $get = null, array $post = null, array $files = null)
    {
        return $this->manageMails('tab_mails', $package, $service, $get, $post, $files);
    }

    /**
     * Client Mail redirection tab
     *
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $get Any GET parameters
     * @param array $post Any POST parameters
     * @param array $files Any FILES parameters
     * @return string The string representing the contents of this tab
     */
    public function tabClientMails($package, $service, array $get = null, array $post = null, array $files = null)
    {
        return $this->manageMails('tab_client_mails', $package, $service, $get, $post, $files);
    }

    /**
     * Handle updating whois information
     *
     * @param string $view The view to use
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $get Any GET parameters
     * @param array $post Any POST parameters
     * @param array $files Any FILES parameters
     * @return string The string representing the contents of this tab
     */
    private function manageWhois($view, $package, $service, array $get = null, array $post = null, array $files = null)
    {
        $this->_logNetim(__FUNCTION__,[$view, $package,$service,$get,$post,$files],"input");

        $vars = new stdClass();
        $domain = $this->getServiceDomain($service);
        $whois_fields = Configure::get('Netim.whois_fields');
        $whois_sections = Configure::get('Netim.whois_sections');
        $error_message="";
        
        // Domain Info
        try
        {
            //get API
            $api = $this->getApi($this->getModuleRow($package->module_row));
            $resDomainInfo = $api->domainInfo($domain);
        }
        catch (NetimAPIException $exception)
        {  
            $this->view = new View("service_error", 'default');
            $this->view->setDefaultView('components' . DS . 'modules' . DS . 'netim' . DS);
            $error_message = Language::_('Netim.!error.domain_loading', true);
            $this->view->set('error_message', $error_message);
            $return = $this->view->fetch();
            $this->_logNetim(__FUNCTION__,$return,"output");
            return $return;
        }
        finally
        {
            $this->_logAPIRequest(__FUNCTION__,$api);
        }

        //we check if contact are specials contacts. If they are we MUST NOT contactOwnerUpdateObj/contactUpdateObj
        $blockChangeOwner = preg_match('/^NETIM4|^TRUST4|^#AUTO#/i',$resDomainInfo->idOwner);
        $blockChangeAdmin = preg_match('/^NETIM4|^TRUST4/i',$resDomainInfo->idAdmin);
        $blockChangeTech = preg_match('/^NETIM4|^TRUST4/i', $resDomainInfo->idTech);
        $blockChangeBilling = preg_match('/^NETIM4|^TRUST4/i', $resDomainInfo->idBilling);

        // Load Owner Contact info
        if (!$blockChangeOwner)
        {
            try
            {
                $resOwner = $api->contactInfoObj($resDomainInfo->idOwner);
            }
            catch (NetimAPIException $exception)
            {
                $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                $this->_logNetim(__FUNCTION__,null,"output");
                return;
            }
            finally
            {
                $this->_logAPIRequest(__FUNCTION__,$api);
            }
        }
		
		
        // Load Admin contact info
        if (!$blockChangeAdmin)
        {
            try
            {
                $resAdmin = $api->contactInfoObj($resDomainInfo->idAdmin);
            }
            catch (NetimAPIException $exception)
            {
                $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                $this->_logNetim(__FUNCTION__,null,"output");
                return;
            }
            finally
            {
                $this->_logAPIRequest(__FUNCTION__,$api);
            }
        }
		
        // Load tech contact info
        if (!$blockChangeTech)
        { 
            try
            {
                $resTech = $api->contactInfoObj($resDomainInfo->idTech);
            }
            catch (NetimAPIException $exception)
            {
                $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                $this->_logNetim(__FUNCTION__,"","output");
                return;
            }
            finally
            {
                $this->_logAPIRequest(__FUNCTION__,$api);
            }
        }
		
        // Load Billing contact info
        if (!$blockChangeBilling)
        {
            try
            {
                $resBilling = $api->contactInfoObj($resDomainInfo->idBilling);
            }
            catch (NetimAPIException $exception)
            {
                $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                $this->_logNetim(__FUNCTION__,null,"output");
                return;
            }
            finally
            {
                $this->_logAPIRequest(__FUNCTION__,$api);
            }
        }

        $pendingOpe = Netim\Database\getAsyncRequest($this->Record, 'domainChangeContact', $domain);
        $pending = null;
        if ($pendingOpe)
        {
            $opes = Netim\Database\getAllAsyncRequests($this->Record);
            foreach($opes as $ope)
            {
                if ($ope->code_ope == 'domainChangeContact' && $ope->domaine == $domain)
                {
                    $pending = $ope->id_ope;
                }
            }
        }

        if(!empty($post) && $pendingOpe)
            $this->Input->setErrors(array('api' =>array('errors' => Language::_('Netim.tab_whois.client_pending', true))));
		
        if (!empty($post) && !$pendingOpe) // is called when clicking on "update" the whois tab
		{
			//$post contains every field from whois_fields
            $post = array_merge(array_intersect_key($post, $whois_fields));
			
			$vars = (object)$post;
			
			if($blockChangeAdmin && $blockChangeTech && $blockChangeBilling)
			{
				$this->Input->setErrors(array('api' =>array('errors' => "All contact are protected contact, you can't edit them")));
				return ;
            }

            if (!$blockChangeAdmin)
                $resAdmin->id = $resDomainInfo->idAdmin;
            if (!$blockChangeTech)
                $resTech->id = $resDomainInfo->idTech;
            if (!$blockChangeBilling)
                $resBilling->id = $resDomainInfo->idBilling;
                
			foreach(array('Admin'=>'Admin', 'Tech'=>'Tech', 'Billing'=>'Billing') as $contactBlesta => $contactNetim)
			{
                // Check that the contact is not a special contact that we shouldn't modify
                if(($contactBlesta === 'Admin' && !$blockChangeAdmin) || ($contactBlesta === 'Tech' && !$blockChangeTech) || ($contactBlesta === 'Billing' && !$blockChangeBilling))
                {
                    $contact = new NormalizedContact($vars->{$contactBlesta.'FirstName'},$vars->{$contactBlesta.'LastName'},$vars->{$contactBlesta.'BodyName'},$vars->{$contactBlesta.'Address1'},
                    $vars->{$contactBlesta.'Address2'},$vars->{$contactBlesta.'PostalCode'}, $vars->{$contactBlesta.'StateProvince'},$vars->{$contactBlesta.'Country'},
                    $vars->{$contactBlesta.'City'}, $vars->{$contactBlesta.'Phone'}, $vars->{$contactBlesta.'EmailAddress'}, 'EN', 0);
                    
                    if(!$contact->equals(${'res'.$contactNetim}))
                    {
                        try
                        {
                            ${'id' . $contactNetim} = $api->contactCreateObj($contact);
                        }
                        catch (NetimAPIException $exception)
                        {
                            $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                            break;
                        }
                        finally
                        {
                            $this->_logAPIRequest(__FUNCTION__,$api);
                        }
                    }
                }
            }
            
            // If any of these contacts changed, update the domain
            if(isset($idAdmin) || isset($idTech) || isset($idBilling))
            {
                try
                {
                    $res = $api->domainChangeContact($domain, $idAdmin ?? $resDomainInfo->idAdmin, $idTech ?? $resDomainInfo->idTech, $idBilling ?? $resDomainInfo->idBilling);

                    if($res->STATUS == "Done")
                    {

                    }
                    else if ($res->STATUS == "Pending")
                    {
                        Netim\Database\addAsyncRequest($this->Record, $res->ID_OPE, "domainChangeContact", $domain);
                        $this->Input->setErrors(array('api' =>array('errors' => Language::_('Netim.tab_whois.error.update_pending', true))));
                    }
                    else if ($res->STATUS == "Failed")
                    {
                        $this->Input->setErrors(array('api' =>array('errors' => $res->MESSAGE)));
                    }
                }
                catch (NetimAPIException $exception)
                {
                    $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                }
                finally
                {
                    $this->_logAPIRequest(__FUNCTION__,$api);
                }
            }

            //New owner object containing the potentially edited values
            if (!$blockChangeOwner)
            {
                $editedOwner = new NormalizedContact($resOwner->getFirstName(), $resOwner->getLastName(), $resOwner->getBodyName(), $vars->RegistrantAddress1, $vars->RegistrantAddress2,
                $vars->RegistrantPostalCode, $vars->RegistrantStateProvince, $vars->RegistrantCountry, $vars->RegistrantCity, $vars->RegistrantPhone, $vars->RegistrantEmailAddress, 'EN', 1);
                if(!$editedOwner->equals($resOwner))
                {
                    try
                    {
                        $res = $api->contactOwnerUpdateObj($resDomainInfo->idOwner, $editedOwner);
                    }
                    catch (NetimAPIException $exception)
                    {
                        $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                    }
                    finally
                    {
                        $this->_logAPIRequest(__FUNCTION__,$api);
                    }
                }
            }
            
            if (!$blockChangeOwner)
            {    
                // To keep displaying name/company
                $vars->RegistrantFirstName = $editedOwner->getFirstName();
                $vars->RegistrantLastName = $editedOwner->getLastName();
                $vars->RegistrantBodyName = $editedOwner->getBodyName();
            }
        }
		else // get called when loading the whois tab
		{
            $contactArray = array();
            if (!$blockChangeOwner)
                $contactArray['Registrant'] = 'Owner';
            if (!$blockChangeAdmin)
                $contactArray['Admin'] = 'Admin';
            if (!$blockChangeTech)
                $contactArray['Tech'] = 'Tech';
            if (!$blockChangeBilling)
                $contactArray['Billing'] = 'Billing';
            
            foreach($contactArray as $contactBlesta => $contactNetim)
			{
				$vars->{$contactBlesta.'FirstName'} = ${'res'.$contactNetim}->getFirstName();
                $vars->{$contactBlesta.'LastName'} = ${'res'.$contactNetim}->getLastName();
                $vars->{$contactBlesta.'BodyName'} = ${'res'.$contactNetim}->getBodyName();
				$vars->{$contactBlesta.'Address1'} = ${'res'.$contactNetim}->getAddress1();
				$vars->{$contactBlesta.'Address2'} = ${'res'.$contactNetim}->getAddress2();
				$vars->{$contactBlesta.'City'} = ${'res'.$contactNetim}->getCity();
				$vars->{$contactBlesta.'StateProvince'} = ${'res'.$contactNetim}->getArea();
				$vars->{$contactBlesta.'PostalCode'} = ${'res'.$contactNetim}->getZipCode();
				$vars->{$contactBlesta.'Country'} = ${'res'.$contactNetim}->getCountry();
				$vars->{$contactBlesta.'Phone'} = ${'res'.$contactNetim}->getPhone();
				$vars->{$contactBlesta.'EmailAddress'} = ${'res'.$contactNetim}->getEmail();
			}
        }
     
        // Process the view
        $this->view = new View($view, 'default');
        Loader::loadHelpers($this, array('Form', 'Html', 'Javascript'));
        $this->view->set('vars', $vars);
        $this->view->set('fields', $this->arrayToModuleFields($whois_fields, null, $vars)->getFields());
        $this->view->set('sections', array('Registrant', 'Admin', 'Tech', 'Billing'));
        $this->view->set('blockChange', array('Registrant' => $blockChangeOwner, 'Admin' => $blockChangeAdmin, 'Billing' => $blockChangeBilling, 'Tech' => $blockChangeTech));
        $this->view->set('blockCountry', array(
                'Registrant' => isset($vars->RegistrantCountry)?$vars->RegistrantCountry:"", 
                'Admin' => isset($vars->AdminCountry)?$vars->AdminCountry:"", 
                'Billing' => isset($vars->BillingCountry)?$vars->BillingCountry:"", 
                'Tech' => isset($vars->TechCountry)?$vars->TechCountry:""
            ));
        $this->view->set('countryCodes', Netim\Core\getCountryCodes($this->Record));
        $this->view->set('stateCodes', Netim\Core\getStateCodes($this->Record));
        $this->view->set('pending', $pending);
        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'netim' . DS);

        $return = $this->view->fetch();

        $this->_logNetim(__FUNCTION__,$return,"output");

        return $return;
    }

    /**
     * Handle updating nameserver information
     *
     * @param string $view The view to use
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $get Any GET parameters
     * @param array $post Any POST parameters
     * @param array $files Any FILES parameters
     * @return string The string representing the contents of this tab
     */
    private function manageNameservers($view, $package, $service, array $get = null, array $post = null, array $files = null)
	{
        $this->_logNetim(__FUNCTION__,[$view,$package,$service,$get,$post,$files],"input");

        $vars = new stdClass();    
        $domain = $this->getServiceDomain($service);
        $error_message ="";
        
        $isPendingOpe = Netim\Database\getAsyncRequest($this->Record, 'domainChangeDNS', $domain);
        $pending = null;
        if ($isPendingOpe)
        {
            $opes = Netim\Database\getAllAsyncRequests($this->Record);
            foreach($opes as $ope)
            {
                if ($ope->code_ope == 'domainChangeDNS' && $ope->domaine == $domain)
                    $pending = $ope->id_ope;
            }
        }

        if(!empty($post) && $isPendingOpe)
            $this->Input->setErrors(array('api' =>array('errors' => Language::_('Netim.tab_nameservers.client_pending', true))));
     
        if (!empty($post) && ! $isPendingOpe) 
		{
            try
            {
                //get API
                $api = $this->getApi($this->getModuleRow($package->module_row));

                $response = $api->domainChangeDNS($domain, $post['ns1'], $post['ns2'], $post['ns3'], $post['ns4'], $post['ns5']);

                if ($response->STATUS == "Pending")
                {
                    Netim\Database\addAsyncRequest($this->Record, $response->ID_OPE, 'domainChangeDNS', $domain);
                    $this->Input->setErrors(array('api' =>array('errors' => Language::_('Netim.tab_nameservers.error.update_pending', true))));
                    $pending = $response->ID_OPE;
                }
                if ($response->STATUS == 'Failed')
                {
                    $this->Input->setErrors(array('api' =>array('errors' => $response->MESSAGE)));
                }
            }
            catch (NetimAPIException $exception)
            {
                $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
            }
            finally
            {
                $this->_logAPIRequest(__FUNCTION__,$api);
            }

            $vars = (object)$post;
            $vars->ns[0] = $post['ns1'];
            $vars->ns[1] = $post['ns2'];
            $vars->ns[2] = $post['ns3'];
            $vars->ns[3] = $post['ns4'];
            $vars->ns[4] = $post['ns5'];
        }
		else
		{
            try
            {
                $api = $this->getApi($this->getModuleRow($package->module_row));
                $domainInfo = $api->domainInfo($domain);
                foreach ($domainInfo->ns as $ns)
				    $vars->ns[] = $ns;
            }
            catch (NetimAPIException $exception)
            {
                $view="service_error";
                $error_message = Language::_('Netim.!error.domain_loading', true);
            }   
            finally
            {
                $this->_logAPIRequest(__FUNCTION__,$api);
            }
        }

        // Process the view
        $this->view = new View($view, 'default');
        Loader::loadHelpers($this, array('Form', 'Html'));
        $this->view->set('vars', $vars);
        $this->view->set('error_message', $error_message);
        $this->view->set('pending', $pending);
        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'netim' . DS);

        $return = $this->view->fetch();
        
        $this->_logNetim(__FUNCTION__,$return,"output");        

        return $return;
    }

    /**
     * Handle dns secure
     *
     * @param string $view The view to use
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $get Any GET parameters
     * @param array $post Any POST parameters
     * @param array $files Any FILES parameters
     * @return string The string representing the contents of this tab
     */
    private function manageDnssec($view, $package, $service, array $get = null, array $post = null, array $files = null)
	{
        $this->_logNetim(__FUNCTION__,[$view,$package,$service,$get,$post,$files],"input");

        $vars = new stdClass();
        $domain = $this->getServiceDomain($service);
        $error_message="";

        $pendingOpe = Netim\Database\getAsyncRequest($this->Record, 'domainSetDNSSec', $domain) || Netim\Database\getAsyncRequest($this->Record, 'domainSetDNSSecExt', $domain);;
        $pending = null;
        if ($pendingOpe)
        {
            $opes = Netim\Database\getAllAsyncRequests($this->Record);
            foreach($opes as $ope)
            {
                if (($ope->code_ope == 'domainSetDNSSec' || $ope->code_ope == 'domainSetDNSSecExt') && $ope->domaine == $domain)
                    $pending = $ope->id_ope;
            }
        }

        if(!empty($post) && $pendingOpe)
            $this->Input->setErrors(array('api' =>array('errors' => Language::_('Netim.tab_dnssec.client_pending', true))));
        
        // Post
        if(!empty($post) && !$pendingOpe)
		{
            if ($post['action'] == 'update' && isset($post['netimDNSSec']))
            {
                try
                {
                    //get API
                    $api = $this->getApi($this->getModuleRow($package->module_row));

                    $res = $api->domainSetDNSSec($domain, $post['netimDNSSec']);

                    if ($res->STATUS == "Pending")
                    {
                        Netim\Database\addAsyncRequest($this->Record, $res->ID_OPE, 'domainSetDNSSec', $domain);
                        $this->Input->setErrors(array('api' =>array('errors' => Language::_('Netim.tab_dnssec.error.update_pending', true))));
                    }
                    else if ($res->STATUS == "Failed")
                    {
                        $this->Input->setErrors(array('api' =>array('errors' => $res->MESSAGE)));
                    }
                }
                catch (NetimAPIException $exception)
                {
                    $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                }
                finally
                {
                    $this->_logAPIRequest(__FUNCTION__,$api);
                }
            }
            else if ($post['action'] == 'update')
            {
                try
                {
                    $api = $this->getApi($this->getModuleRow($package->module_row));
                    $res = $api->domainSetDNSSecExt($domain, array(array('key' => $post['dnssec_key'], 'type' => $post['dnssec_type'], 'digest' => $post['dnssec_digest'])), $post['dnssec_flags'], $post['dnssec_protocol'], $post['dnssec_algo'], $post['dnssec_pubkey']);
                    if ($res->STATUS == "Pending")
                    {
                        Netim\Database\addAsyncRequest($this->Record, $res->ID_OPE, 'domainSetDNSSecExt', $domain);
                        $pending = $res->ID_OPE;
                    }
                    else if ($res->STATUS == "Failed")
                    {
                        $this->Input->setErrors(array('api' =>array('errors' => $res->MESSAGE)));
                    }
                }
                catch(Exception $exception)
                {
                    $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                }
                finally
                {
                    $this->_logAPIRequest(__FUNCTION__,$api);
                }
            }

            $vars = (object)$post;
        }

        // Get Domain Info
        try
        {
            $api = $this->getApi($this->getModuleRow($package->module_row));
            $tldInfo = $api->domainTldInfo($this->getTld($domain));
            $domainInfo = $api->domainInfo($domain);
        }
        catch (NetimAPIException $exception)
        {
            $view = "service_error";
            $error_message = Language::_('Netim.!error.domain_loading', true);
        }
        finally
        {
            $this->_logAPIRequest(__FUNCTION__,$api);
        }

        if(!$tldInfo->HasDnsSec)
        {
            $view = "service_error";
            $error_message = Language::_('Netim.!error.service_unavailable', true);
        }

        // Process the view
        $this->view = new View($view, 'default');
        Loader::loadHelpers($this, array('Form', 'Html'));
        $this->view->set('vars', $vars);
        $this->view->set('isSigned', $domainInfo->IsSigned);
        if ($domainInfo->IsSigned)
            $this->view->set('dnssecArray', $domainInfo->DNSSEC);
        $this->view->set('error_message', $error_message);
        $this->view->set('pending', $pending);
        $this->view->set('hasDNS4Service', $domainInfo->HasDNS4Service);
        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'netim' . DS);

        $return = $this->view->fetch();

        $this->_logNetim(__FUNCTION__,$return,"output");

        return $this->view->fetch();
    }

    /**
     * Handle updating settings
     *
     * @param string $view The view to use
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $get Any GET parameters
     * @param array $post Any POST parameters
     * @param array $files Any FILES parameters
     * @return string The string representing the contents of this tab
     */
    private function manageSettings($view, $package, $service, array $get = null, array $post = null, array $files = null)
	{
        $this->_logNetim(__FUNCTION__,[$view,$package,$service,$get,$post,$files],"input");

        Loader::loadModels($this, array("Services"));
        $vars = new stdClass();
        $domain = $this->getServiceDomain($service);
        $error_message="";

        // Gathering TLD infos
        try
        {
            //get API
            $api = $this->getApi($this->getModuleRow($package->module_row));
            $tldInfo = $api->domainTldinfo($this->getTld($domain));

            // Availability of the 3 features for the corresponding tld
            $registrarLockEnabled = ($tldInfo->HasRegistrarLock == '1');
            $whoisPrivacyEnabled  = ($this->featureServiceEnabled('id_protection',$service) == '1');
            $eppCodeEnabled       = ($tldInfo->HasEppCode       == '1');
            $membershipEnabled    = ($this->getTld($domain) == 'xxx');
            
        }
        catch (NetimAPIException $exception)
        {
            $this->view = new View("service_error", 'default');
            $error_message = Language::_('Netim.!error.domain_loading', true);
            $this->view->set('error_message', $error_message);
            $this->view->setDefaultView('components' . DS . 'modules' . DS . 'netim' . DS);
            $return = $this->view->fetch();
            $this->_logNetim(__FUNCTION__,$return,"output");
            return $return;
        }    
        finally
        {
            $this->_logAPIRequest(__FUNCTION__,$api);
        }
        
        // Post
        if(!empty($post))
		{
            if(isset($post['save']))
            {
                if(isset($post['registrar_lock']) && $registrarLockEnabled)
                {
                    try
                    {
                        $response = $api->domainSetRegistrarLock($domain, $post['registrar_lock']);
                    }
                    catch (NetimAPIException $exception)
                    {
                        $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                    }
                    finally
                    {
                        $this->_logAPIRequest(__FUNCTION__,$api);
                    }
                }

                if(isset($post['whois_privacy']) && $whoisPrivacyEnabled)
                {
                    try
                    {
                        $response = $api->domainSetWhoisPrivacy($domain, $post['whois_privacy']);
                    }
                    catch (NetimAPIException $exception)
                    {
                        $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                    }
                    finally
                    {
                        $this->_logAPIRequest(__FUNCTION__,$api);
                    }
                }
            }

            if(isset($post['request_epp']) && $eppCodeEnabled)
            {
                if(isset($post['field_request_epp']))
                {
                    try
                    {
                        $response = $api->domainAuthID($domain, 1);
                    }
                    catch (NetimAPIException $exception)
                    {
                        $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                    }
                    finally
                    {
                        $this->_logAPIRequest(__FUNCTION__,$api);
                    }
                }
            }
            

            if(isset($post['delete']) && $view == "tab_settings")
            {  
                try
                {
                    $response = $api->domainDelete($domain, 'NOW');
                    if ($response->STATUS == 'Done')
                    {
                        if (!$this->Services->edit($service->id, ['status' => "canceled"],true,true));
                            $this->Input->setErrors(array('api' =>array('errors' => "Domain has been deleted at the Registrar but the cancelation in Blesta failed")));
                    }
                    if ($response->STATUS == 'Pending')
                    {
                        Netim\Database\addAsyncRequest($this->Record, $response->ID_OPE, 'domainDelete', $domain);
                        $this->Input->setErrors(array('api' =>array('errors' => "Operation is pending with id " . $response->ID_OPE)));
                    }
                    if ($response->STATUS == 'Failed')
                    {
                        $this->Input->setErrors(array('api' =>array('errors' => "Operation failed : " . $response->MESSAGE)));
                    }
                }
                catch (NetimAPIException $exception)
                {
                    $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                }
                finally
                {
                    $this->_logAPIRequest(__FUNCTION__,$api);
                }
            }

            if(isset($post['membership']))
            {
                try
                {
                    $response = $api->domainSetMembership($domain,$post['token']);
                    
                    if($response->STATUS=="Pending")
                    {
                        Netim\Database\addAsyncRequest($response->ID_OPE, 'domainSetmembership', $domain, $params["token"]);
                        $values["error"] = str_replace("[ID_OPE]",$response->ID_OPE, $_ADMINLANG['netim']['request']['pending_admin_change']);
                    }
        
                    if($response->STATUS=="Failed")
                        $values["error"] = $response->MESSAGE;
                }
                catch(NetimAPIException $fault)
                {
                    $values["error"] = $fault->getMessage();
                }
                finally
                {
                    $this->_logAPIRequest(__FUNCTION__,$api);
                }
            }

            $vars = (object)$post;
        }
		else
		{
            try
            {
                $response = $api->domainInfo($domain);
                
                $vars->registrar_lock = $response->domainIsLock;
                $vars->whois_privacy = $response->whoisPrivacy;
                $vars->field_request_epp = '0';
            }
            catch (NetimAPIException $exception)
            {
                $view = "service_error";
                $error_message = Language::_('Netim.!error.domain_loading', true); 
            }
            finally
            {
                $this->_logAPIRequest(__FUNCTION__,$api);
            }
        } 

        // Delete Enabled ?        
        if($service->status == "canceled")
        {
            $deleteEnabled = 0;
        }
        else
        {
            $deleteEnabled = 1;
        }
                
        // Process the view
        $this->view = new View($view, 'default');
        Loader::loadHelpers($this, array('Form', 'Html'));
        $this->view->set('vars', $vars);
        $this->view->set('error_message', $error_message);
        $this->view->set('enabledFields', array('registrarLockEnabled' => $registrarLockEnabled, 'whoisPrivacyEnabled' => $whoisPrivacyEnabled, 'eppCodeEnabled' => $eppCodeEnabled,'deleteEnabled' => $deleteEnabled , 'membershipEnabled' => $membershipEnabled));
        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'netim' . DS);

        $return = $this->view->fetch();

        $this->_logNetim(__FUNCTION__,$return,"output");

        return $return;
    }

    private function manageDnsRecords($view, $package, $service, array $get = null, array $post = null, array $files = null)
    {
        $this->_logNetim(__FUNCTION__,[$view,$package,$service,$get,$post,$files],"input");

        $vars = new stdClass();
        $domain = $this->getServiceDomain($service);
        $recordTypes = array('A','AAAA','MX','CNAME','TXT');
        $webRedirectEnabled = (isset($package->meta->web));
        if($webRedirectEnabled)
            $recordTypes = array_merge($recordTypes, ['URL']);
        $error_message="";

        // Post
        if(!empty($post))
		{
            if($post['action'] == 'deleteDnsRecord')
            {
                $subdomain = rtrim(str_replace($domain, '', $post['host']), '.');
                $value = rtrim($post['value']);
                
                if($post['type'] == "URL")
                {
                    try {
                        //get API
                        $api = $this->getApi($this->getModuleRow($package->module_row));

                        $api->domainWebFwdDelete($post['fqdn']);
                    }
                    catch (NetimAPIException $exception) {
                        $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                    }
                    finally {
                        $this->_logAPIRequest(__FUNCTION__,$api);
                    }
                }
                else
                {
                    try {
                        //get API
                        $api = $this->getApi($this->getModuleRow($package->module_row));

                        $api->domainZoneDelete($domain, $subdomain, $post['type'], $value);
                    }
                    catch (NetimAPIException $exception) {
                        $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                    }
                    finally {
                        $this->_logAPIRequest(__FUNCTION__,$api);
                    }
                }
            }
            else if ($post['action'] == 'addDnsRecord')
            {
                $type = $recordTypes[$post['record_type']]; 
                $value = rtrim($post['value']);
                $subdomain = rtrim($post['subdomain']);
                $fqdn = "$subdomain.$domain";
                $structOperationResponse = array("service"=>null,"protocol"=>null,"ttl"=>null,"priority"=>$post['priority'],"weight"=>null,"port"=>null);
                if($type == "URL")
                {

                    $url = parse_url($value);
                    $structOptionsFwd = array();
                    $structOptionsFwd["header"] = $post['redirection']?302:301;
                    $structOptionsFwd["protocol"] = $url["scheme"];
                    $structOptionsFwd["title"] = "";
                    $structOptionsFwd["parking"] = "";

                    try {
                        //get API
                        $api = $this->getApi($this->getModuleRow($package->module_row));

                        $api->domainWebFwdCreate($fqdn, $value, 'DIRECT', $structOptionsFwd);
                    }
                    catch (NetimAPIException $exception) {
                        $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                    }
                    finally {
                        $this->_logAPIRequest(__FUNCTION__,$api);
                    }
                }
                else
                {
                    try {
                        //get API
                        $api = $this->getApi($this->getModuleRow($package->module_row));

                        $api->domainZoneCreate($domain, $subdomain, $type, $value, $structOperationResponse);
                    }
                    catch (NetimAPIException $exception) {
                        $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                    }
                    finally {
                        $this->_logAPIRequest(__FUNCTION__,$api);
                    }
                }
            }
        }

        // Get Domain Info
        try
        {                        
            //get API
            $api = $this->getApi($this->getModuleRow($package->module_row));

            $domainInfo = $api->domainInfo($domain);
        }
        catch (NetimAPIException $exception)
        {
            $view = "service_error";
            $error_message = Language::_('Netim.!error.domain_loading', true);
        }
        finally
        {
            $this->_logAPIRequest(__FUNCTION__,$api);
        }

        if($domainInfo)
        {
            // Get service Info
            if(!$domainInfo->HasDNS4Service)
            {
                $view = "service_error";
                $error_message = Language::_('Netim.!error.service_unavailable_dns', true);
            }
            else
            {
                try
                {
                    $webFwd = $api->queryWebFwdList($domain);
                    $zoneList = $api->queryZoneList($domain);

                    foreach($webFwd as $fwd)
                    {
                        foreach($zoneList as $zone)
                        {
                            if($zone->host == $fwd->FQDN)
                                unset($zoneList[array_search($zone, $zoneList)]);
                        }
                    }
                }
                catch (NetimAPIException $exception)
                {
                    $view="service_error";
                    $error_message =Language::_('Netim.!error.domain_loading', true);
                }
                finally
                {
                    $this->_logAPIRequest(__FUNCTION__,$api);
                }
            }
        }
        
        $vars = array('webFwd' => $webFwd, 'zoneList' => $zoneList, 'recordTypes' => $recordTypes);

        // Process the view
        $this->view = new View($view, 'default');
        Loader::loadHelpers($this, array('Form', 'Html'));
        $this->view->set('domain', $domain);
        $this->view->set('vars', $vars);
        $this->view->set('post', $post);
        $this->view->set('error_message', $error_message);
        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'netim' . DS);

        $return = $this->view->fetch();

        $this->_logNetim(__FUNCTION__,$return,"output");

        return $return;
    }

    /**
     * Handle updating host information
     *
     * @param string $view The view to use
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $get Any GET parameters
     * @param array $post Any POST parameters
     * @param array $files Any FILES parameters
     * @return string The string representing the contents of this tab
     */
    private function manageHosts($view, $package, $service, array $get = null, array $post = null, array $files = null)
	{
        $this->_logNetim(__FUNCTION__,[$view,$package,$service,$get,$post,$files],"input");

        $vars = array();
        $domain = $this->getServiceDomain($service);
        $error_message="";

        // Post
        if (!empty($post)) 
		{
            if(!empty($post['ipv4']))
                $ipv4s = explode(';', $post['ipv4']);
            else 
                $ipv4s = array();
            if(!empty($post['ipv6'])) 
                $ipv6s = explode(';', $post['ipv6']);
            else 
                $ipv6s = array();

            if($post['action'] == "updateHost")
            {
                try
                {
                    $api = $this->getApi($this->getModuleRow($package->module_row));
                    $res = $api->hostUpdate($post['hostname'], $ipv4s, $ipv6s);
                }
                catch (NetimAPIException $exception) {
                    $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                }
                finally {
                    $this->_logAPIRequest(__FUNCTION__,$api);
                }
            }
            else if($post['action'] == "addHost")
            {
                try {
                    $api = $this->getApi($this->getModuleRow($package->module_row));
                    $res = $api->hostCreate($post['hostname'] . "." . $domain, $ipv4s, $ipv6s);
                }
                catch (NetimAPIException $exception) {
                    $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                }
                finally {
                    $this->_logAPIRequest(__FUNCTION__,$api);
                }
            }
            else if ($post['action'] == "deleteHost")
            {
                try {
                    $api = $this->getApi($this->getModuleRow($package->module_row));
                    $res = $api->hostDelete($post['hostname']);
                }
                catch (NetimAPIException $exception) {
                    $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                }
                finally {
                    $this->_logAPIRequest(__FUNCTION__,$api);
                }
            }
        }
        
        // Load
        try
        {
            $api = $this->getApi($this->getModuleRow($package->module_row));
            $vars['hosts'] = $api->queryHostList("*".$domain);
            $opeInfo = $api->QueryOpeList($this->getTld($domain));
        }
        catch (NetimAPIException $exception)
        {
            $view = "service_error";
            $error_message = Language::_('Netim.!error.domain_loading', true);
        }
        finally
        {
            $this->_logAPIRequest(__FUNCTION__,$api);
        }

        if($opeInfo->hostCreate==0)
        {
            $view = "service_error";
            $error_message = Language::_('Netim.!error.service_unavailable', true);
        }
        
        // Process the view
        $this->view = new View($view, 'default');
        Loader::loadHelpers($this, array('Form', 'Html'));
        $this->view->set('vars', $vars);
        $this->view->set('post', $post);
        $this->view->set('domain', $domain);
        $this->view->set('error_message', $error_message);
        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'netim' . DS);

        $return = $this->view->fetch();

        $this->_logNetim(__FUNCTION__,$return,"output");

        return $return;
    }

    /**
     * Handle updating mail redirections
     *
     * @param string $view The view to use
     * @param stdClass $package A stdClass object representing the current package
     * @param stdClass $service A stdClass object representing the current service
     * @param array $get Any GET parameters
     * @param array $post Any POST parameters
     * @param array $files Any FILES parameters
     * @return string The string representing the contents of this tab
     */
    private function manageMails($view, $package, $service, array $get = null, array $post = null, array $files = null)
    {
        $this->_logNetim(__FUNCTION__,[$view,$package,$service,$get,$post,$files],"input");

        $vars = array();
        $domain = $domain = $this->getServiceDomain($service);;
        $error_message="";
        
        // Post
        if (!empty($post)) 
		{
            if ($post['action'] == 'addMail')
            {
                try {    
                    $api = $this->getApi($this->getModuleRow($package->module_row));
                    $api->domainMailFwdCreate($post['source']."@".$domain, $post['destination']);
                }
                catch (NetimAPIException $exception) {
                    $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                }
                finally {
                    $this->_logAPIRequest(__FUNCTION__,$api);
                }
            }
            else if ($post['action'] == 'deleteMail')
            {
                try {
                    $api = $this->getApi($this->getModuleRow($package->module_row));
                    $api->domainMailFwdDelete($post['source']);
                }
                catch (NetimAPIException $exception) {
                    $this->Input->setErrors(array('api' =>array('errors' => $exception->getMessage())));
                }
                finally {
                    $this->_logAPIRequest(__FUNCTION__,$api);
                }
            }
        }

        // Get Domain Info
        try
        {
            $api = $this->getApi($this->getModuleRow($package->module_row));
            $domainInfo = $api->domainInfo($domain);
        }
        catch (NetimAPIException $exception)
        {
            $view = "service_error";
            $error_message = Language::_('Netim.!error.domain_loading', true);
        }
        finally
        {
            $this->_logAPIRequest(__FUNCTION__,$api);
        }

        if($domainInfo)
        {
            // Get service Info
            if(!$domainInfo->HasDNS4Service)
            {
                $view = "service_error";
                $error_message = Language::_('Netim.!error.service_unavailable_dns', true);
            }
            else
            {
                try
                {
                    $vars['mails'] = $api->queryMailFwdList($domain);
                }
                catch (NetimAPIException $exception)
                {
                    $view="service_error";
                    $error_message =Language::_('Netim.!error.domain_loading', true);
                }
                finally
                {
                    $this->_logAPIRequest(__FUNCTION__,$api);
                }
            }
        }

        // Process the view
        $this->view = new View($view, 'default');
        Loader::loadHelpers($this, array('Form', 'Html'));
        $this->view->set('vars', $vars);
        $this->view->set('post', $post);
        $this->view->set('domain', $domain);
        $this->view->set('error_message', $error_message);
        $this->view->setDefaultView('components' . DS . 'modules' . DS . 'netim' . DS);

        $return = $this->view->fetch();

        $this->_logNetim(__FUNCTION__,$return,"output");

        return $return;
    }

    /**
     * Returns all fields to display to an admin attempting to add a service with the module
     *
     * @param stdClass $package A stdClass object representing the selected package
     * @param $vars stdClass A stdClass object representing a set of post fields
     * @return ModuleFields A ModuleFields object, containg the fields to render as
     *  well as any additional HTML markup to include
     */
    public function buildDomainModuleFields($fields, $vars)
    {
        
        $user_lang = Configure::get('Blesta.language');
        $module_fields=$this->arrayToModuleFields($fields, null, $vars);

        // Add additional domain fields
        $tld = $this->getTld($vars->domain);
        $additional_fields_settings=Netim\Core\getAdditionalFields($tld);
        if($additional_fields_settings)
        {
            //Load translations
            include(__DIR__ . DS . 'additional_fields' . DS . $user_lang.'.inc.php');     

            $additional_fields=array();
            $display_rules=false;
            foreach($additional_fields_settings as $additional_field) 
            {
                $key = $additional_field->Key;

                $label=Netim\Core\translate($additional_field->Name,$user_lang);
                $description = isset($additional_field->Description) ? Netim\Core\translate($additional_field->Description,$user_lang): "";
                $type = $additional_field->Type;

                $items = (isset($additional_field->Items) ? $additional_field->Items : (object)array());
                $options = (isset($additional_field->Options) ? $additional_field->Options : (object)array());

                $field_id = $key . '_id';
                $field_attributes['id'] = $field_id;
                $field_label = $module_fields->label($label, $field_id); // Create a label
                
                switch($type)
                {
                    case "display":
                        //Blesta doesn't support display
                        //In order to display the rules, we transform it as a checkbox for agreement.

                        //Is there a display for rules ?
                        if($key=="rules")
                            $display_rules=true;

                        // We don't take into account display for information
                        if($key=="infos") 
                            break;                              

                        if(isset($options->Agreement) && !empty($options->Agreement))
                            $description.=" ".Netim\Core\translate($options->Agreement,$user_lang);
                        else if($key=="rules")
                            $description.=" ".$netim_agreement_rules;
                        else if($key=="notices")
                            $description.=" ".$netim_agreement_rules2;

                        $option_id = $field_id . '_1';
                        $option_label = $module_fields->label($description, $option_id); //Create an item label
                        $checked=isset($vars->{$key}) && $vars->{$key}=="checked";
                        $field=$module_fields->fieldCheckbox($key,"checked",$checked,['id' => $option_id],$option_label);  //Create a field of type checkbox
                        $field_label->attach($field); // Attach the field to the label
                        $module_fields->setField($field_label); // Add the label to the fields
                        break;

                    case "text":
                        $value=(isset($vars->{$key})?$vars->{$key}:"");
                        $field=$module_fields->fieldText($key,$value,$field_attributes); //Create a field of type text
                        $field_label->attach($field); // Attach the field to the label
                        if(!empty($description))
                        {
                            $field_tooltip=$module_fields->tooltip($description); //Create a tooltip
                            $field_label->attach($field_tooltip); // Attach the tooltip to the label
                        }
                        $module_fields->setField($field_label); // Add the label to the fields
                        break;
                       
                    case "checkbox":
                        // We don't take into account agreement if there is a rule (will acts as agreement...)
                        if($key=="agreement" && $display_rules) 
                            break;

                        $option_id = $field_id . '_1';
                        $option_label = $module_fields->label($description, $option_id); //Create an item label
                        $checked=isset($vars->{$key}) && $vars->{$key}=="checked";
                        $field=$module_fields->fieldCheckbox($key,"checked",$checked,['id' => $option_id],$option_label);  //Create a field of type checkbox
                        $field_label->attach($field); // Attach the field to the label
                        $module_fields->setField($field_label); // Add the label to the fields
                        break;
                       
                    case "radio":
                        $i = 1;

                        foreach($items as $item)
                        {
                            $option_id = $field_id . '_' . $i++;
                            $desc=Netim\Core\translate($item->Description,$user_lang);
                            $option_label = $module_fields->label($desc, $option_id); //Create an item label
                            $checked=isset($vars->{$key}) && $vars->{$key}==$item->Value;
                            $field=$module_fields->fieldRadio($key,$item->Value,$checked,['id' => $option_id],$option_label);  //Create a field of type radio
                            $field_label->attach($field); // Attach the field to the item label
                        }

                        if(!empty($description))
                        {
                            $field_tooltip=$module_fields->tooltip($description); //Create a tooltip
                            $field_label->attach($field_tooltip); // Attach the tooltip to the label
                        }
                        $module_fields->setField($field_label); // Add the label to the fields               
                        break;
                       
                    case "dropdown":
                        $options=array(""=>"---"); // Empty value to validate a mandatory choice

                        foreach($items as $item)
                        {
                            $desc=Netim\Core\translate($item->Description);
                            $options[$item->Value]=$desc;
                        }

                        $selected=$vars->{$key};   
                        $field=$module_fields->fieldSelect($key,$options,$selected,$field_attributes); //Create a field of type select
                        $field_label->attach($field); // Attach the field to the label
                        if(!empty($description))
                        {
                            $field_tooltip=$module_fields->tooltip($description); //Create a tooltip
                            $field_label->attach($field_tooltip); // Attach the tooltip to the label
                        }
                        $module_fields->setField($field_label); // Add the label to the fields
                        break;  
                }
            }             
        }     
        
        return $module_fields;
    }

    public function getAdminAddFields($package, $vars = null)
    {
        // Are we in the admin > Add service feature ?    
        if(!isset($vars->step))
            return $this->getClientAddFields($package, $vars); // No, then display client fields

        // yes, display specific fields to add the service
        if ($package->meta->type == 'domain') {

            // Set default name servers
            $this->setDefaultNS($package, $vars);
            
            // Choose a domain action
            $fields['domainAction'] = [
                'label' => Language::_('Netim.domain.ActionType', true),
                'type' => 'radio',
                'value' => '0',
                'options' => [
                    '0' => Language::_('Netim.domain.ActionRegister', true),
                    '1' => Language::_('Netim.domain.ActionTransfer', true),
                ],
            ];

            
            if (isset($vars->domainAction) && $vars->domainAction==1) {
                // Handle transfer request    
                $fields = array_merge($fields,Configure::get('Netim.transfer_fields'),Configure::get('Netim.nameserver_fields'));

                // Hide transfer because it is an hidden info
                $fields['transfer']['type'] = 'hidden';
                $fields['transfer']['label'] = null;
                $fields['transfer']['options'] = 'true';

                // Manage EPP code
                //$row = $this->getModuleRow(empty($vars->module_row_id)?$vars->module_row_id:$package->module_row);
                try 
                {
                    $api = $this->getApi($this->getModuleRow($package->module_row));
                    $tld = $api->domainTldInfo($this->getTld($vars->domain));
                    if($tld->HasEppCode==0)
                        unset($fields['transfer_key']);
                }
                catch (NetimAPIException $exception)
                {}

                $module_fields=$this->arrayToModuleFields($fields, null, $vars);
            } else {
               // Handle registration request 
                $fields = array_merge($fields,Configure::get('Netim.domain_fields'),Configure::get('Netim.nameserver_fields'));
                $module_fields = $this->buildDomainModuleFields($fields, $vars);
            }
        }

        // Check the default action
        if(!isset($vars->domainAction))
        {
            $module_fields->setHtml("
            <script type=\"text/javascript\">
                $(document).ready(function() {
                    $('#domainAction_id_0').prop('checked', true);
                 });
            </script>");
        }

        return (isset($module_fields) ? $module_fields : new ModuleFields());
    }

    /**
     * Returns all fields to display to an admin attempting to edit a service with the module
     *
     * @param stdClass $package A stdClass object representing the selected package
     * @param $vars stdClass A stdClass object representing a set of post fields
     * @return ModuleFields A ModuleFields object, containg the fields to render
     *  as well as any additional HTML markup to include
     */
    public function getAdminEditFields($package, $vars = null)
    {
        return new ModuleFields();
    }

    /**
     * Returns all fields to display to a client attempting to add a service with the module
     *
     * @param stdClass $package A stdClass object representing the selected package
     * @param $vars stdClass A stdClass object representing a set of post fields
     * @return ModuleFields A ModuleFields object, containg the fields to render
     *  as well as any additional HTML markup to include
     */
    public function getClientAddFields($package, $vars = null)
    {
        // Set default name servers
        $this->setDefaultNS($package, $vars);

        if (isset($vars->transfer)) {
            // Handle transfer request    
            $fields = array_merge(Configure::get('Netim.transfer_fields'),Configure::get('Netim.nameserver_fields'));

            // We should already have the domain name don't make editable
            $fields['domain']['type'] = 'hidden';
            $fields['domain']['label'] = null;

            // Hide transfer because it is an hidden info
            $fields['transfer']['type'] = 'hidden';
            $fields['transfer']['label'] = null;
            $fields['transfer']['options'] = 'true';

            // Manage EPP code
            try 
            {
                $api = $this->getApi($this->getModuleRow($package->module_row));
                $tld = $api->domainTldInfo($this->getTld($vars->domain));
                if($tld->HasEppCode==0)
                    unset($fields['transfer_key']);
            }
            catch (NetimAPIException $exception){
            }
            finally {
                $this->_logAPIRequest(__FUNCTION__,$api); 
            }

            return $this->arrayToModuleFields($fields, null, $vars);
        } else {
            // Handle domain registration
            $fields = array_merge(Configure::get('Netim.domain_fields'),Configure::get('Netim.nameserver_fields'));
            $fields['domain']['type'] = 'hidden';
            $fields['domain']['label'] = null;
            $module_fields = $this->buildDomainModuleFields($fields, $vars);             
            return $module_fields;  
        }
    }

    /**
     * Returns all fields to display to a client attempting to edit a service with the module
     *
     * @param stdClass $package A stdClass object representing the selected package
     * @param $vars stdClass A stdClass object representing a set of post fields
     * @return ModuleFields A ModuleFields object, containg the fields to render
     *  as well as any additional HTML markup to include
     */
    public function getClientEditFields($package, $vars = null)
    {
        return new ModuleFields();
    }
    
     /**
     * Get a list of the TLDs supported by the registrar module
     *
     * @param int $module_row_id The ID of the module row to fetch for the current module
     * @return array A list of all TLDs supported by the registrar module
     */
    public function getTlds($module_row_id = null)
    {
        //Getting all tlds that cannot be sold on blesta
        $rejected_tlds = Configure::get('Netim.rejected_tlds');

        //Getting all TLDs from netim
		$TldData = $this->getNetimRawTldData();
        if(isset($TldData["error"]))
        {
            $this->_logNetim(__FUNCTION__,$TldData["error"],"output");
            return [];
        }

        $tlds = array();
        for($i = 0; $i < count($TldData['extension']); $i++)
        {
            $tld = ((array) $TldData['extension'][$i]);
            array_push($tlds, ".".$tld['@attributes']['name']);
        }

        //Removing rejected tlds
        foreach ($rejected_tlds as $rejected_tld)
        {
            if(($key = array_search($rejected_tld, $tlds))!== false)
                unset($tlds[$key]);
        }

        return $tlds;
    }

	/**
	* Performs an availability lookup on the given domain
	*
	* @param string $domain The domain to lookup
	* @return bool True if available, false otherwise
	*/
    public function checkAvailability($domain, $module_row_id = null)   
    {
        include(dirname(__FILE__) . "/config/netim.php");

        if($netim_check_debug)
            $this->_logNetim(__FUNCTION__,$domain,"input");

        $res = null;
        try {
            $api = $this->getApi($this->getModuleRow($package->module_row));
            $res = $api->domainCheck($domain)[0];
        }
        catch (NetimAPIException $exception) {
        }
        finally {
            if($netim_check_debug)
                $this->_logAPIRequest(__FUNCTION__,$api); 
        }

        // We consider that if reason is not empty, the domain is not available (premium not supported, and so on)
        $result = (is_object($res) && $res->result == 'AVAILABLE' && empty($res->reason));
        if($netim_check_debug)
            $this->_logNetim(__FUNCTION__,($result ? "AVAILABLE FOR REGISTRATION":"NOT AVAILABLE"),"output");
    
        return $result;
    }

    /**
     * Verifies that the provided domain name is available for transfer
     *
     * @param string $domain The domain to lookup
     * @param int $module_row_id The ID of the module row to fetch for the current module
     * @return bool True if the domain is available for transfer, false otherwise
     */
    public function checkTransferAvailability($domain, $module_row_id = null)
    {
        include(dirname(__FILE__) . "/config/netim.php");

        if($netim_check_debug)
            $this->_logNetim(__FUNCTION__,$domain,"input");
		
        //Special case for acceptance tests by Netim
        if(preg_match("#blesta-transfer#",$domain))
            return true;

        try {
            $api = $this->getApi($this->getModuleRow($package->module_row));
            $res = $api->domainCheck($domain);
            $res = $res[0];
        }
        catch (NetimAPIException $exception) {
        }finally {
            if($netim_check_debug)
                $this->_logAPIRequest(__FUNCTION__,$api);   
        }

        // We consider that if reason is not empty, the domain is not available (premium not supported, and so on)
        $result = (is_object($res) && $res->result == 'NOT AVAILABLE' && empty($res->reason));
        if($netim_check_debug)
            $this->_logNetim(__FUNCTION__,($result ? "AVAILABLE FOR TRANSFER":"NOT AVAILABLE"),"output");
    
        return $result;
    }

    /**
	* Cron task
	*
	*/
    public function cron($key) 
    {
        Loader::loadModels($this, array("Services", "Packages"));
        $content="";

        switch ($key) {
            case "netim_async": {
                
                $requests = Netim\Database\getAllAsyncRequests($this->Record);
                foreach($requests as $request)
                {
                    $content.="Processing ".$request->domaine." with request #".$request->id_ope."\n";

                    //Get the service id from database
                    $service_id = Netim\Core\GetDomainID($this->Record,$request->domaine);
                    if(!$service_id)
                    {
                        $content.=">> ERROR Service_id not found\n";

                        //Remove the domain from async requests
                        Netim\Database\removeAsyncRequest($this->Record,$request->id_ope);

                        continue;  
                    }
                    else
                        $content.=">> Service_id = $service_id\n";

                    //Load the service object
                    $service = $this->Services->get($service_id);
                    if(!$service)
                    {
                        $content.=">> ERROR Service not loaded\n";
                        continue;  
                    }
                    
                    //Load the corresponding module row
                    $row = $this->getModuleRow($service->module_row_id);
                    if(!$row)
                    {
                        $content.=">> ERROR Module Row not loaded\n";
                        continue;  
                    }

                    //Query request info
                    try
                    {  
                        $api = $this->getApi($row);
                        $ope = $api->queryOpe($request->id_ope);
                        $content.=">> Result = $ope->STATUS\n";
                    }
                    catch (NetimAPIException $exception)
                    {
                        $content.=">> API Error (queryOpe): ".$exception->getMessage()."\n";
                        $content.=">> PASS\n";
                        continue;
                    }
                    finally
                    {
                        $this->_logAPIRequest(__FUNCTION__,$api);
                    }

                    if($ope->STATUS=="Done")
				    {
                        //Query domain info
                        try
                        {  
                            $domainInfo = $api->domainInfo($request->domaine);
                        }
                        catch (NetimAPIException $exception)
                        {
                            $content.=">> API Error (domainInfo): ".$exception->getMessage()."\n";
                            $content.=">> PASS\n";
                            continue;
                        }
                        finally
                        {
                            $this->_logAPIRequest(__FUNCTION__,$api);
                        }

                        //Process successful request      
                        switch($request->code_ope)
                        {
                            case "domainCreate":
                                
                                $vars=['date_added' => $domainInfo->dateCreate,
                                'date_renews' => $domainInfo->dateExpiration,
                                'status' => 'active'
                                ];

                                if($this->Services->edit($service_id, $vars,false,true)) // 3rd param must be false to send notification
                                {    
                                    $content.=">> Domain successfully activated\n";
                                    
                                    //Remove the domain from async requests
                                    Netim\Database\removeAsyncRequest($this->Record,$request->id_ope);
                                }     

                                break; 

                            case "domainRenew":
                                
                                $vars=['date_renews' => $domainInfo->dateExpiration,
                                'status' => 'active'
                                ];

                                if($this->Services->edit($service_id, $vars,true,false))
                                {    
                                    $content.=">> Domain successfully renewed\n";
                                    
                                    //Remove the domain from async requests
                                    Netim\Database\removeAsyncRequest($this->Record,$request->id_ope);
                                }     
                               
                                break;
                            
                            case "domainTransferTrade":        
                            case "domainTransferIn":
                                // Blesta doesn't manage a special pricing and special duration for domain transfer.
                                // Netim doesn't manage duration for domain transfer too => An explicit renewal is needed in some cases
                                // If transfer is paid then one year is included in the pricing. Additional renewal is needed according to the term ordered
                                
                                // Is an additional renew needed ?
                                try
                                {    
                                    $prices = $api->queryDomainPrice($request->domaine);
                                    $cost = $prices->Fee4Transfer;      
                                }
                                catch (NetimAPIException $exception)
                                {
                                    $content.=">> API Error (queryDomainPrice): ".$exception->getMessage()."\n";
                                    $content.=">> PASS\n";
                                    break;
                                }
                                finally
                                {
                                    $this->_logAPIRequest(__FUNCTION__,$api);
                                }

                                if($cost>0)
                                    $additional_duration = $service->package_pricing->term-1;
                                else if($cost==0)
                                    $additional_duration = $service->package_pricing->term;
                                    
                                if($additional_duration>0)
                                {  
                                    $content.= ">> Additional renewal is required in order to extend the domain with $additional_duration year(s)\n";
                                    try
                                    {    
                                        $result  = $api->domainRenew($request->domaine, $additional_duration);
                                        
                                        if($result->STATUS=="Pending" || $result->STATUS=="Done")
                                        {
                                            // We manage both cases asynchronously to update the expiration accordingly after the activation
                                            Netim\Database\addAsyncRequest($this->Record, $result->ID_OPE, 'domainRenew', $request->domaine);
                                            $content.= ">> Renewal is requested with number #: ".$result->ID_OPE." and is under processing\n";
                                        }
                            
                                        if($result->STATUS=="Failed")
                                            $content.= ">> Renewal was requested with number #: ".$result->ID_OPE." and Failed. Ensure to synchronize the domain expiration according to the term ordered by the client\n";
                                    }
                                    catch (NetimAPIException $exception)
                                    {
                                        $content.= ">> API Error (domainRenew): ".$exception->getMessage()."\n";
                                        $content.= ">> Renewal was requested and Failed. Ensure to synchronize the domain expiration according to the term ordered by the client\n";
                                    } 
                                    finally
                                    {
                                        $this->_logAPIRequest(__FUNCTION__,$api);
                                    }
                                }  

                                // Activation of the domain
                                $vars=['date_renews' => $domainInfo->dateExpiration,
                                'status' => 'active'
                                ];

                                if($this->Services->edit($service_id, $vars,false,true)) // 3rd param must be false to send notification
                                {    
                                    $content.=">> Domain successfully transferred\n";
                                    
                                    //Remove the domain from async requests
                                    Netim\Database\removeAsyncRequest($this->Record,$request->id_ope);
                                }  
                                   
                                break; 
                            case "domainChangeDNS":
                            case "domainSetDNSSec":
                            case "domainSetDNSSecExt":
                            case "domainChangeContact":
                                //Remove the domain from async requests
                                Netim\Database\removeAsyncRequest($this->Record,$request->id_ope);
                                break;
                            case "domainDelete":
                                //Remove the domain from async requests
                                Netim\Database\removeAsyncRequest($this->Record,$request->id_ope);

                                 // Cancellation of the domain
                                $edit = $this->Services->edit($serviceId, ['status' => "canceled"],true,true);
                                if($edit)
                                    $content.=">> Successfully deleted ". $request->domaine."\n";
                                else
                                    $content.=">> ERROR Deletion of domain " . $request->domaine . " failed\n";
                                break;
                            default:
                                $content.=">> PASS\n";        
                        }
                    }
                    else if($ope->STATUS=="Failed" || $ope->STATUS=="Cancelled")
                    {
                        //Process failed or cancelled request     
                        switch($request->code_ope)
                        {
                            case "domainRenew":
                                //Query domain info
                                try
                                {
                                    $domainInfo = $api->domainInfo($request->domaine);

                                    //Set back the initial expiration date
                                    $edit=$this->Services->edit($service->id, ['date_renews' => $domainInfo->dateExpiration],true,false);
                                    if($edit)
                                    {
                                        $content.=">> Set Date renews back to ".$domainInfo->dateExpiration."\n";
                                     
                                        //Remove the domain from async requests
                                        Netim\Database\removeAsyncRequest($this->Record,$request->id_ope); 
                                    }             
                                    else
                                        $content.=">> ERROR Update of Date renews failed\n"; 
                                }
                                catch (NetimAPIException $exception)
                                {
                                    $content.=">> API Error (domainInfo): ".$exception->getMessage()."\n";
                                    $content.=">> PASS\n";
                                }
                                finally
                                {
                                    $this->_logAPIRequest(__FUNCTION__,$api);
                                }
                                break;
                            case "domainCreate":
                                Netim\Database\addError($this->Record, $ope->MESSAGE, 'domainCreate', $request->domaine, $request->id_ope);
                                Netim\Database\removeAsyncRequest($this->Record,$request->id_ope);
                                break;
                            case "domainTransferIn":
                                Netim\Database\addError($this->Record, $ope->MESSAGE, 'domainTransferIn', $request->domaine, $request->id_ope);
                                Netim\Database\removeAsyncRequest($this->Record,$request->id_ope);
                                break;
                            case "domainTransferTrade":
                                Netim\Database\addError($this->Record, $ope->MESSAGE, 'domainTransferTrade', $request->domaine, $request->id_ope);
                                Netim\Database\removeAsyncRequest($this->Record,$request->id_ope);
                                break;
                                
                            default:
                                //Remove the domain from async requests
                                Netim\Database\removeAsyncRequest($this->Record,$request->id_ope);     
                        } 
                    }
                    else
                        $content.=">> PASS\n"; 
                }
            break;
            }     
        }

        $this->Log2Cron($content);  
    }

    /**
	* Write logs in the cron log
	*
	*/
    private function Log2Cron($strlog) 
    {
        //Cron() function should return a string content to be written in log.
        //in core/Automation/Tasks/Task/Module.php process()
        //Update requested to Blesta team
        //Alternative: throw an exception
        if(!empty($strlog))
        {
            $strlog.="\n\nBelow stack trace is not intended and it is not an error:";
            throw new Exception($strlog);       
        }       
    }

    /**
     * Initializes the NETIM and returns an instance of that object
     *
     * @param string $user The user to connect as
     * @param string $key The key to use when connecting
     * @param bool $sandbox Whether or not to process in sandbox mode (for testing)
     * @return APISoap The ClientAPI instance
     */
    private function getApi($row, $user ="", $key ="", $sandbox="")
    {
        Loader::load( __DIR__ . DS . 'api' . DS . 'APISoap.php');
        
        if($row) {
            $user = $row->meta->user;
            $key = $row->meta->key;
            $sandbox = $row->meta->sandbox;
        }
        $api = null;

        if(!is_bool($sandbox))
            $sandbox=($sandbox=="true");
        
        if (!class_exists('SoapClient'))
            throw new Exception("SOAP functions are not available<br>Please check PHP pre-requisites at https://support.netim.com/en/docs/blesta/installation", 1);
  
		try {
            $api = new Netim\APISoap($user, $key, $sandbox);

            //Don't update setVersion() call, it used by NETIM to track and follow module usage
            $api->setVersion("BLESTA=".BLESTA_VERSION.",PLUGIN=".self::MODULE_VERSION);
            
            //Increase synchronization value
            $api->sessionSetPreference("syncDelay",25);
        } 
        catch (\Exception $fault) {
            $this->_logNetim(__FUNCTION__,$fault->GetMessage());
            throw new NetimAPIException($fault->GetMessage(), 1);
        }

        return $api;
    }

    /**
     * Get User Language
     */

    public function getUserLanguage()
    {
        return Configure::get('Blesta.language');
    }
	
    /**
	* Logs a module message
	*
	* @param $url string The function / location
	* @param $log mixed The message to be logged
    * @param $pur string input or output
	*/
    private function _logNetim($url, $log, $put = "input")
    {
        include(dirname(__FILE__) . "/config/netim.php");

        if(!$netim_debug)
            return;

        if(!is_string($log))
        {
            $t = print_r($log,true);
            $log3 = $t;
        }
        else
            $log3 = $log;

        if($put == "input")
            $log2 = "\nParam :\n".$log3;
        else if($put == "output")
            $log2 = "\nReturn :\n".$log3;

        $result = "Success";

        try 
        {
            $this->log($url, $log2, $put, $result);
        }
        catch (Exception $e) 
        {}
    }

    /**
    * Logs the API request
    *
    * @param $fn string The function were the call was done
    * @param $api object The Netim ClientAPI object
    * @return bool Result
    */
    private function _logAPIRequest($fn,$api)
    {
        include(dirname(__FILE__) . "/config/netim.php");

		if(!$netim_api_debug)
			return;

        if(is_null($api))
            return;

        // Result
        $result=empty($api->getLastError());
        if($result)
            $response=$api->getLastResponse();
        else
            $response=$api->getLastError();    
        
        // Input
        $logI = "\nAPI function : ".$api->getLastRequestFunction();
        if(!is_string($api->getLastRequestParams()))
            $t = print_r($api->getLastRequestParams(),true);
        else
            $t = $api->getLastRequestParams();
        $logI.= "\n\nAPI Request:\n".$t;

        // Output
        $logO = "\nAPI function : ".$api->getLastRequestFunction();
        if(!is_string($response))
            $t = print_r($response,true);
        else
            $t = $response;
        $logO.="\n\nAPI Response:\n".$t;

        try {
            $this->log($fn, $logI, 'input', $result);
            $this->log($fn, $logO, 'output', $result);
        }
        catch (Exception $e){
        }
    }

    /**
	* Logs the API request
    *
    * @param $fn string The function were the call was done
    * @param $api object The Netim ClientAPI object
    * @param $apifn string The function called by the API
    * @param $apifn_param The param for the $apifn
	*/
    private function logAPIRequest($fn,$api,$apifn = null,$apifn_param = null)
    {
        include(dirname(__FILE__) . "/config/netim.php");

		if(!$netim_api_debug)
			return;

        $result=empty($api->getLastError());

        if($result)
            $response=$api->getLastResponse();
        else
            $response=$api->getLastError();    
        
        if($apifn != null && $apifn_param != null)
        {
            $logI = "\nAPI function : ".$apifn;
            if(!is_string($apifn_param)){
                $t = print_r($apifn_param,true);
            }
            else{
                $t = $apifn_param;
            }
            $logI.= "\n\nAPI Request:\n".$t;
            $logO = "\nAPI function : ".$apifn;
            $logO.="\n\nAPI Response:\n".var_export($response,true);
        }
        else
        {
            $logI="\n\nAPI Request:\n".var_export($api->getLastRequestParams(),true);
            $logO="\n\nAPI Response:\n".var_export($response,true);
        }

        try 
        {
            $this->log($fn, $logI, 'input', $result);
            $this->log($fn, $logO, 'output', $result);
        }
        catch (Exception $e) 
        {}
    }

	/**
	* Returns the TLD of the given domain
	*
	* @param string $domain The domain to return the TLD from
	* @return string The TLD of the domain
	*/
    private function getTld($domain)
    {
        $domain = strtolower($domain);

        /*
        $tlds = $this->getTlds();      
        foreach ($tlds as $tld) {
            if (substr($domain, -strlen($tld)) == $tld) {
                return $tld;
            }
        }
        */
        return substr(strstr($domain, '.'),1,strlen($domain));
    }

	/**
	* Set default ns
	* 
	* WARNING: args by reference!!!
	*
	* @param Object $package
	* @param Object $vars
	*
	* MODIFY $package and $vars!
	*/
	private function setDefaultNS(&$package, &$vars)
	{
		if (!isset($vars->ns1) && isset($package->meta->ns))
		{
			$i=1;
			foreach ($package->meta->ns as $ns)
			{
				$vars->{'ns' . $i++} = $ns;
			}	
		}
	}
	
	/**
	* Set unspecified checkboxes
	*
	* WARNING: args by reference!!!
	*
	* @param Object $vars
	*
	* MODIFY $var!
	*/
	private function setUnspecifiedCheckboxes(&$vars)
	{
		if (empty($vars['sandbox']))
		{
			$vars['sandbox'] = 'false';
		}
	}


    /**
     * Get a list of raw TLD data
     *
     * @param int $module_row_id The ID of the module row to fetch for the current module
     *
     * @return stdClass The response from the cache or API
     */

    private function getNetimRawTldData(){

        // Fetch the TLDs results from the cache, if they exist
        $cache = Cache::fetchCache(
            'netim_tlds',
            Configure::get('Blesta.company_id') . DS . 'modules' . DS . 'netim' . DS
        );
       
        if ($cache)
            $xmlobj = unserialize(base64_decode($cache));

        // Get remote tlds list
        if (!isset($xmlobj)) {

            if(!ini_get('allow_url_fopen'))
                return ["error" => "allow_url_fopen PHP setting is not enabled<br>Please check PHP pre-requisites at https://support.netim.com/en/docs/blesta/installation"];

            if (!function_exists('simplexml_load_string'))
                return ["error" => "Simple XML functions are not available<br>Please check PHP pre-requisites at https://support.netim.com/en/docs/blesta/installation"];

            $xmlfile= Configure::get('Netim.tlds_url');
		    $xmlstr=file_get_contents($xmlfile);

            if(empty($xmlstr))
            {
                $error = print_r(error_get_last(),true);
                return ['error' => "Unable to fetch tld information : $error"];
            }

            if(!preg_match("#<tld-list>#",$xmlstr))
                return ['error' => "Unable to fetch tld information : Bad xml format"];

            $xml = simplexml_load_string($xmlstr);
            if($xml===false)
                return ['error' => "Unable to fetch tld information : Bad xml format"];

            $xmlobj = json_decode(json_encode((array) $xml),1);

            // Save all TLDs form netim in cache
            if (Configure::get('Caching.on') && is_writable(CACHEDIR)) {
                try {
                    Cache::writeCache(
                        'netim_tlds',
                        base64_encode(serialize($xmlobj)),
                        strtotime(Configure::get('Blesta.cache_length')) - time(),
                        Configure::get('Blesta.company_id') . DS . 'modules' . DS . 'netim' . DS
                    );
                } catch (Exception $e) {
                    // Write to cache failed, so disable caching
                    Configure::set('Caching.on', false);
                }
            }
        }

        return $xmlobj;
    }

    /**
     * Get a list of raw TLD pricing data
     *
     * @param int $module_row_id The ID of the module row to fetch for the current module
     *
     * @return stdClass The response from the cache or API
     */

    private function getNetimRawPricingData($module_row_id = null){

        // Fetch the pricing results from the cache, if they exist
        $cache = Cache::fetchCache(
            'netim_pricing',
            Configure::get('Blesta.company_id') . DS . 'modules' . DS . 'netim' . DS
        );
        
        if ($cache) {
            $xmlobj = unserialize(base64_decode($cache));
        }

        if (!isset($xmlobj)){

            if (!function_exists('curl_init'))
                return ["error" => "Curl functions are not available<br>Please check PHP pre-requisites at https://support.netim.com/en/docs/blesta/installation"];

            if (!function_exists('simplexml_load_string'))
                return ["error" => "Simple XML functions are not available<br>Please check PHP pre-requisites at https://support.netim.com/en/docs/blesta/installation"];

            $row = $this->getModuleRow($module_row_id);
            if (!$row) {
                $rows = $this->getModuleRows();
                if (isset($rows[0])) {
                    $row = $rows[0];
                }
                unset($rows);
            }
      
            // Curl request to get data
            $ch = curl_init();
            if($row->meta->sandbox)
                curl_setopt($ch, CURLOPT_URL,Configure::get('Netim.pricing_url_ote'));
            else
                curl_setopt($ch, CURLOPT_URL,Configure::get('Netim.pricing_url'));
            curl_setopt($ch, CURLOPT_POST, 1);
            curl_setopt($ch, CURLOPT_POSTFIELDS, http_build_query(array('LOGIN' => $row->meta->user,"PWD" => $row->meta->key)));
            curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
            $xmlstr = curl_exec($ch);
            curl_close ($ch);

            if(empty($xmlstr))
            {
                $error = print_r(error_get_last(),true);
                return ['error' => "Unable to fetch pricing information : $error"];
            }

            if(preg_match("#error#",$xmlstr))
                return ['error' => "Unable to fetch pricing information : $xmlstr"];
 
            if(!preg_match("#<pricing>#",$xmlstr))
                return ['error' => "Unable to fetch pricing information : Bad xml format"];
 
            $xml = simplexml_load_string($xmlstr);
            if($xml===false)
                return ['error' => "Unable to fetch tld information : Bad xml format"];

            $xmlobj = json_decode(json_encode((array) $xml),1);

            // Save all TLDs form netim in cache
            if (Configure::get('Caching.on') && is_writable(CACHEDIR)) {
                try {
                    Cache::writeCache(
                        'netim_pricing',
                        base64_encode(serialize($xmlobj)),
                        strtotime(Configure::get('Blesta.cache_length')) - time(),
                        Configure::get('Blesta.company_id') . DS . 'modules' . DS . 'netim' . DS
                    );
                } catch (Exception $e) {
                    // Write to cache failed, so disable caching
                    Configure::set('Caching.on', false);
                }
            }
        }

        return $xmlobj;
    }

    /**
     * Get a list of the TLD prices
     *
     * @param int $module_row_id The ID of the module row to fetch for the current module
     * @return array A list of all TLDs and their pricing
     *    [tld => [currency => [year# => ['register' => price, 'transfer' => price, 'renew' => price]]]]
     */
    
    public function getTldPricing($module_row_id = null)
    {
        return $this->getFilteredTldPricing($module_row_id);
    }

    /**
     * Get a filtered list of the TLD prices
     *
     * @param int $module_row_id The ID of the module row to fetch for the current module
     * @param array $filters A list of criteria by which to filter fetched pricings including but not limited to:
     *
     *  - tlds A list of tlds for which to fetch pricings
     *  - currencies A list of currencies for which to fetch pricings
     *  - terms A list of terms for which to fetch pricings
     * @return array A list of all TLDs and their pricing
     *    [tld => [currency => [year# => ['register' => price, 'transfer' => price, 'renew' => price]]]]
     */

    public function getFilteredTldPricing($module_row_id = null,$filters = [])
     {
        $this->_logNetim(__FUNCTION__,[$module_row_id,$filters],"input");

        // Get all currencies
        Loader::loadModels($this, ['Currencies']);
        $reseller_currency="EUR";
        
        $currencies = [];
        $company_currencies = $this->Currencies->getAll(Configure::get('Blesta.company_id'));
        foreach ($company_currencies as $currency) {
            $currencies[$currency->code] = $currency;
        }

        // Download TLD data
		$TldData = $this->getNetimRawTldData();
        if(isset($TldData["error"]))
        {
            $this->_logNetim(__FUNCTION__,$TldData["error"],"output");
            throw new Exception($TldData["error"], 1);
            return false;
        }

        $PricingData = $this->getNetimRawPricingData($module_row_id);
        if(isset($PricingData["error"]))
        {
            $this->_logNetim(__FUNCTION__,$PricingData["error"],"output");
            throw new Exception($PricingData["error"], 1);
            return false;
        }

        // Format pricing
        $tld_yearly_prices = [];

        for($i = 0; $i < count($TldData['extension']); $i++)
        {
            // Get the TLD
            $tld = ((array) $TldData['extension'][$i]);
            $tld_name = $tld['@attributes']['name'];
            $tld_name2 = "." . ltrim($tld_name,".");

            // Filter by 'tlds'
            if (isset($filters['tlds']) && !in_array($tld_name2, $filters['tlds']))
                continue;

            // Get the TLD pricing
            $ind_p = 1;
            for($j = 0; $j < count($PricingData['domain']); $j++)
            {
                $tld_p = $PricingData['domain'][$j];
                if($tld_p['extension'] == $tld_name){
                    $ind_p = $j;
                    break;
                }
            }

            if (!isset($tld_yearly_prices[$tld_name2]))
                $tld_yearly_prices[$tld_name2] = [];
           
            // Get the TLD price
            $last_year = [1,10];

            foreach(range(1,(int) $last_year[1]) as $year){

                // Filter by 'terms'
                if (isset($filters['terms']) && !in_array($year, $filters['terms']))
                    continue;

                foreach($currencies as $currency){

                    // Filter by 'currencies'
                    if (isset($filters['currencies']) && !in_array($currency->code, $filters['currencies'])) {
                        continue;
                    }

                    $price_c = (int)$tld_p['domainCreate'];
                    $price_t = (int)$tld_p['domainTransferIn'];
                    $price_r = (int)$tld_p['domainRenew'];

                    // Set Price entries with default value
                    if (!isset($tld_yearly_prices[$tld_name2][$currency->code]))
                        $tld_yearly_prices[$tld_name2][$currency->code] = [];

                    if (!isset($tld_yearly_prices[$tld_name2][$currency->code][$year]))
                        $tld_yearly_prices[$tld_name2][$currency->code][$year] = [
                            'register' => null,
                            'transfer' => null,
                            'renew' => null
                        ];

                    // Set Registration price
                    if($reseller_currency == $currency->code)
                        $tld_yearly_prices[$tld_name2][$currency->code][$year]['register'] = $price_c * $year;
                    else
                        $tld_yearly_prices[$tld_name2][$currency->code][$year]['register'] = $this->Currencies->convert(
                            $price_c * $year,
                            $reseller_currency,
                            $currency->code,
                            Configure::get('Blesta.company_id')
                        );
   
                    // Set Transfer price
                    if($price_t > 0){
                        if($year == 1)
                            $price_t2 = $price_t;
                        else
                            $price_t2 = $price_t + $price_r * ($year-1);
                    }else if($price_t == 0)
                        $price_t2 = $price_r * $year;
                    
                    if($reseller_currency == $currency->code)
                        $tld_yearly_prices[$tld_name2][$currency->code][$year]['transfer'] = $price_t2;
                    else   
                        $tld_yearly_prices[$tld_name2][$currency->code][$year]['transfer'] = $this->Currencies->convert(
                            $price_t2,
                            $reseller_currency,
                            $currency->code,
                            Configure::get('Blesta.company_id')
                        );

                    // Set Renew price
                    if($reseller_currency == $currency->code)
                        $tld_yearly_prices[$tld_name2][$currency->code][$year]['renew'] = $price_r * $year;
                    else    
                        $tld_yearly_prices[$tld_name2][$currency->code][$year]['renew'] = $this->Currencies->convert(
                            $price_r * $year,
                            $reseller_currency,
                            $currency->code,
                            Configure::get('Blesta.company_id')
                        );
                }
            }
        }

        $this->_logNetim(__FUNCTION__,$tld_yearly_prices,"output");

        return $tld_yearly_prices;
    }

    /**
     * Gets the domain expiration date
     *
     * @param stdClass $service The service belonging to the domain to lookup
     * @param string $format The format to return the expiration date in
     * @return string The domain expiration date in UTC time in the given format
     * @see Services::get()
     */
    
    public function getExpirationDate($service, $format = 'Y-m-d H:i:s')
    {
        $domain = $this->getServiceDomain($service);

        if($service->status == "pending" || $service->status == "canceled"){
            return null;
        }

        // Get API
        $module_row_id = $service->module_row_id ?? null;
        $row = $this->getModuleRow($module_row_id);

        try
        {
            $api = $this->getApi($row); 
            $domainInfo = $api->domainInfo($this->getServiceDomain($service)); 
        }
        catch(NetimAPIException $fault)
        {
            return null;
        }
        finally
        {
            $this->_logAPIRequest(__FUNCTION__,$api);
        }

        $date = date($format, strtotime($domainInfo->dateExpiration));
        return $date;
    }

    

    /**
     * Gets the domain name from the given service
     *
     * @param stdClass $service The service from which to extract the domain name
     * @return string The domain name associated with the service
     * @see Services::get()
     */
    
    public function getServiceDomain($service)
    {
        if (isset($service->fields)) {
            foreach ($service->fields as $service_field) {
                if ($service_field->key == 'domain') {
                    return $service_field->value;
                }
            }
        }

        return $this->getServiceName($service);
    }
    
}
?>